import glob
import os
import random
import re
from pathlib import Path
from typing import Sequence

import questionary
import rich
from rich.panel import Panel
from rich.style import Style
from tqdm import tqdm

from everyvoice.config.type_definitions import DatasetTextRepresentation
from everyvoice.text.utils import guess_graphemes_in_text, guess_ipa_phones_in_text
from everyvoice.utils import (
    generic_xsv_filelist_reader,
    lower,
    nfc_normalize,
    read_festival,
    slugify,
)
from everyvoice.wizard import TEXT_CONFIG_FILENAME_PREFIX, Step, StepNames, Tour
from everyvoice.wizard.prompts import (
    CUSTOM_QUESTIONARY_STYLE,
    get_response_from_menu_prompt,
)
from everyvoice.wizard.utils import (
    apply_automatic_text_conversions,
    read_unknown_tabular_filelist,
    rename_unknown_headers,
)
from everyvoice.wizard.validators import validate_path

# WAVS & FILELIST


class DatasetNameStep(Step):
    DEFAULT_NAME = StepNames.dataset_name_step

    def prompt(self):
        return input(
            "What would you like to call this dataset? This is needed because EveryVoice lets you train models with multiple sources of data. Please choose a name that distinguishes this data source, e.g. 'john-english' or 'maria-spanish' or something similarly descriptive: "
        )

    def validate(self, response):
        if len(response) == 0:
            print("Sorry, your dataset needs a name.")
            return False
        slug = slugify(response)
        if not slug == response:
            print(
                f"Sorry, your name: '{response}' is not valid, since it will be used to create a file and special characters are not permitted in filenames. Please re-type something like {slug} instead."
            )
            return False
        return True

    def effect(self):
        print(
            f"Great! The Configuration Wizard 🧙 finished the configuration for your dataset named '{self.response}'"
        )


class DatasetPermissionStep(Step):
    DEFAULT_NAME = StepNames.dataset_permission_step
    choices = (
        "No, I don't have permission to use this data.",
        "Yes, I do have permission to use this data.",
    )

    def prompt(self):
        prompt_text = """Do you have permission to use this data to build a TTS model? It is unethical to build a TTS model of a speaker without their knowledge or permission and there can be serious consequences for doing so."""
        return get_response_from_menu_prompt(
            prompt_text=prompt_text,
            choices=self.choices,
        )

    def validate(self, response):
        return response in self.choices

    def effect(self):
        if self.state[StepNames.dataset_permission_step.value].startswith("No"):
            print("OK, we'll ask you to choose another dataset then!")
            self.children = []
            del self.root.state[self.state_subset]


class WavsDirStep(Step):
    DEFAULT_NAME = StepNames.wavs_dir_step

    def prompt(self):
        return questionary.path(
            "Where are your audio files?",
            style=CUSTOM_QUESTIONARY_STYLE,
            only_directories=True,
        ).unsafe_ask()

    def sanitize_input(self, response):
        response = super().sanitize_input(response)
        return response.strip()

    def validate(self, response) -> bool:
        valid_path = validate_path(response, is_dir=True, exists=True)
        if not valid_path:
            return False
        path_expanded = Path(response).expanduser()
        glob_iter = glob.iglob(os.path.join(path_expanded, "**/*.wav"), recursive=True)
        contains_wavs = next(glob_iter, None) is not None
        if not contains_wavs:
            print(
                f"Sorry, no .wav files were found in '{path_expanded}'. Please choose a directory with audio files."
            )
        return valid_path and contains_wavs


class SampleRateConfigStep(Step):
    DEFAULT_NAME = StepNames.sample_rate_config_step

    def prompt(self):
        return questionary.text(
            "What is the sample rate (in Hertz) of your data?",
            style=CUSTOM_QUESTIONARY_STYLE,
        ).unsafe_ask()

    def validate(self, response):
        try:
            self.response = int(response)
            if self.response < 100 or float(response) != self.response:
                print(
                    f"{response} is not a valid sample rate. Please enter an integer representing the sample rate in Hertz of your data."
                )
                return False
            return True
        except ValueError:
            print(
                f"{response} is not a valid sample rate. Please enter an integer representing the sample rate in Hertz of your data."
            )
            return False


class FilelistStep(Step):
    DEFAULT_NAME = StepNames.filelist_step

    def prompt(self):
        return questionary.path(
            "Where is your data filelist?", style=CUSTOM_QUESTIONARY_STYLE
        ).unsafe_ask()

    def sanitize_input(self, response):
        response = super().sanitize_input(response)
        return response.strip()

    def validate(self, response) -> bool:
        return validate_path(response, is_file=True, exists=True)


class FilelistFormatStep(Step):
    DEFAULT_NAME = StepNames.filelist_format_step
    separators = {"psv": "|", "tsv": "\t", "csv": ","}

    def prompt(self):
        return get_response_from_menu_prompt(
            prompt_text="Select which format your filelist is in:",
            choices=("psv", "tsv", "csv", "festival"),
            multi=False,
            search=False,
            return_indices=False,
        )

    def looks_like_sv(self, file_type, separator) -> bool:
        assert self.state
        filelist_path = self.state.get(StepNames.filelist_step)
        initial_records = read_unknown_tabular_filelist(
            filelist_path, delimiter=separator, record_limit=10
        )

        column_count = len(initial_records[0])
        if column_count < 2:
            print(
                f"File '{filelist_path}' does not look like a '{file_type}' file: no record separator found on header line."
            )
            return False

        for i, record in enumerate(initial_records):
            if len(record) != column_count:
                print(
                    f"File '{filelist_path}' does not look like a '{file_type}' file: the {i}th record has a different number of fields than the header row."
                )
                return False

        return True

    def validate(self, response):
        if response == "festival":
            filelist_path = self.state.get(StepNames.filelist_step)
            try:
                _ = read_festival(filelist_path, 10)
                return True
            except ValueError:
                print(f"File '{filelist_path}' is not in the festival format.")
                return False

        separator = self.separators.get(response, None)
        if separator:
            return self.looks_like_sv(response, separator)

        assert False and "the above code covers all valid cases"  # pragma: no cover

    def effect(self):
        """
        This effect occurs after the filelist format is selected. The Tour.state should now
        have both the filelist path and the filelist data type, so we parse the data and add
        the data to the tour state. We then inspect the headers for the data and add steps to
        select the basename and text if they are not already specified in the headers.
        """
        file_type = self.state.get(StepNames.filelist_format_step)
        filelist_path = self.state.get(StepNames.filelist_step)
        if not isinstance(filelist_path, Path):
            filelist_path = Path(filelist_path).expanduser()
        if file_type == "csv":
            self.state["filelist_delimiter"] = ","
        elif file_type == "psv":
            self.state["filelist_delimiter"] = "|"
        elif file_type == "tsv":
            self.state["filelist_delimiter"] = "\t"
        else:
            self.state["filelist_delimiter"] = None
        if file_type == "festival":
            filelist_data_dict = read_festival(filelist_path)
            self.state["filelist_headers"] = list(filelist_data_dict[0].keys())
            self.state["filelist_data"] = filelist_data_dict
        else:
            self.state["filelist_data_list"] = read_unknown_tabular_filelist(
                filelist_path, delimiter=self.state.get("filelist_delimiter")
            )
            self.state["filelist_headers"] = list(self.state["filelist_data_list"][0])
            standard_header_found = False
            self.state["selected_headers"] = []
            try:
                text_index = self.state["filelist_headers"].index(
                    "text"
                )  # call it text until FilelistTextRepresentationStep
                self.state["selected_headers"].append(text_index)
                standard_header_found = True
            except ValueError:
                self.tour.add_step(
                    HeaderStep(
                        name=StepNames.text_header_step,
                        prompt_text="Which column contains the [bold blue]text?",
                        header_name="text",
                        state_subset=self.state_subset,
                    ),
                    self,
                )
            try:
                basename_index = self.state["filelist_headers"].index("basename")
                self.state["selected_headers"].append(basename_index)
                standard_header_found = True
            except ValueError:
                self.tour.add_step(
                    HeaderStep(
                        name=StepNames.basename_header_step,
                        prompt_text="Which column contains the [bold blue]basenames?",
                        header_name="basename",
                        state_subset=self.state_subset,
                    ),
                    self,
                )
            if not standard_header_found:
                self.tour.add_step(
                    HasHeaderLineStep(state_subset=self.state_subset), self
                )


class ValidateWavsStep(Step):
    DEFAULT_NAME = StepNames.validate_wavs_step

    def wav_file_early_validation(self) -> int:
        """Look for missing wav files and return the error count"""
        assert self.state is not None  # fixes mypy errors
        wavs_dir = Path(self.state[StepNames.wavs_dir_step])
        files_not_found = []
        MAX_SAMPLES = 1000
        filelist_data = self.state["filelist_data"]
        file_list_size = len(filelist_data)
        sample: Sequence[int]
        if file_list_size > MAX_SAMPLES:
            print(
                f"Checking a sample of {MAX_SAMPLES} of your audio files to make sure they are present."
            )
            sampled_text = " sampled"
            sample = sorted(random.sample(range(file_list_size), MAX_SAMPLES))
        else:
            print("Checking if all your audio files are present.")
            sampled_text = ""
            sample = range(file_list_size)
        for item in sample:
            record = filelist_data[item]  # +1 to skip past header
            wav_basename = record["basename"]
            wav_filename = wavs_dir / (
                wav_basename + ("" if wav_basename.endswith(".wav") else ".wav")
            )
            if not wav_filename.exists():
                files_not_found.append(wav_filename)
        if files_not_found:
            n = len(files_not_found)
            if n == 1:
                print(
                    f"Warning: wav file '{files_not_found[0]}' was not found, please check your filelist."
                )
            else:
                print(
                    f"Warning: {n}{sampled_text} wav files were not found, including '{files_not_found[0]}' and '{files_not_found[1]}'.\nPlease check your wavs directory '{wavs_dir}' and your filelist."
                )
            return n
        print(f"Great! All{sampled_text} audio files found in directory '{wavs_dir}'.")
        return 0

    def prompt(self):
        error_count = self.wav_file_early_validation()
        if error_count:
            return get_response_from_menu_prompt(
                title="Do you want to pick a different wavs directory?",
                choices=(
                    "Yes",
                    "No, I will fix my audio basenames or add missing audio files later.",
                ),
            )
        else:
            return "OK"

    def validate(self, response):
        return response[:3] in ("OK", "Yes", "No,")

    def effect(self):
        if self.response == "Yes":
            self.tour.add_steps(
                [
                    WavsDirStep(state_subset=self.state_subset),
                    ValidateWavsStep(state_subset=self.state_subset),
                ],
                self,
            )
        elif self.response.startswith("No"):
            rich.print(
                Panel(
                    "Continuing despite missing audio files. Make sure you fix your filelist later or add missing audio files, otherwise entries in your filelist with missing audio files will be skipped during preprocessing and therefore be ignored during training.",
                    title="Missing audio files",
                    border_style=Style(color="#EF1010"),
                )
            )


class FilelistTextRepresentationStep(Step):
    DEFAULT_NAME = StepNames.filelist_text_representation_step
    text_representation_options = tuple(x.value for x in DatasetTextRepresentation)

    def prompt(self):
        return get_response_from_menu_prompt(
            prompt_text=f"Which representation is your text in? Choose '{DatasetTextRepresentation.ipa_phones.value}' if your text data only uses International Phonetic Alphabet characters (punctuation is also OK). Choose '{DatasetTextRepresentation.arpabet}' if your text data uses all ARPABET (punctuation is OK). Choose '{DatasetTextRepresentation.characters}' otherwise.",
            choices=self.text_representation_options,
            multi=False,
            search=False,
            return_indices=False,
        )

    def validate(self, response):
        return response in self.text_representation_options

    def effect(self):
        # Apply the text representation level as the new alias for text:
        for i, header in enumerate(self.state["filelist_headers"]):
            if header == "text":
                self.state["filelist_headers"][i] = self.response


# HEADER SELECTION


class HeaderStep(Step):
    DEFAULT_NAME = StepNames.text_header_step

    def __init__(self, name: str, prompt_text: str, header_name: str, **kwargs):
        super(HeaderStep, self).__init__(name=name, **kwargs)
        self.prompt_text = prompt_text
        self.header_name = header_name

    def prompt(self):
        selected_headers = self.state.get("selected_headers", [])
        choice_indices = [
            x
            for x in range(len(self.state["filelist_headers"]))
            if x not in selected_headers
        ]
        choices = [
            f"{x}: {self.state['filelist_data_list'][0][x]}" for x in choice_indices
        ]
        response = get_response_from_menu_prompt(
            prompt_text=self.prompt_text,
            choices=choices,
            multi=False,
            search=False,
            return_indices=True,
        )
        return choice_indices[response]

    def validate(self, response):
        return isinstance(response, int)

    def effect(self):
        # Rename the filelist header with the standard header name
        if "selected_headers" not in self.state:
            self.state["selected_headers"] = []
        self.state["selected_headers"].append(self.response)
        self.state["filelist_headers"][self.response] = self.header_name


class LanguageHeaderStep(HeaderStep):
    def effect(self):
        # Rename the filelist header with the standard header name
        if "selected_headers" not in self.state:
            self.state["selected_headers"] = []
        self.state["selected_headers"].append(self.response)
        self.state["filelist_headers"][self.response] = self.header_name
        # Rename unselected headers to unknown:
        self.state["filelist_headers"] = rename_unknown_headers(
            self.state["filelist_headers"]
        )
        # re-parse data:
        reload_filelist_data_as_dict(self.state)
        # apply automatic conversions
        self.state["model_target_training_text_representation"] = (
            apply_automatic_text_conversions(
                self.state["filelist_data"],
                self.state[StepNames.filelist_text_representation_step],
            )
        )


class HasHeaderLineStep(Step):
    """Check if the data set has a header line, and insert one if not.

    Subsequent steps can systematically assume that self.state["filelist_data_list"][0]
    is the header row."""

    DEFAULT_NAME = StepNames.data_has_header_line_step
    choices = ("no", "yes")

    def prompt(self):
        prompt_text = (
            "Your filelist does not have the standard headers. The first row is:\n"
            + self.state["filelist_delimiter"].join(self.state["filelist_data_list"][0])
            + "\nIs this line a header row?"
        )
        return get_response_from_menu_prompt(
            prompt_text=prompt_text,
            choices=self.choices,
        )

    def validate(self, response):
        return response in self.choices

    def effect(self):
        if self.state[StepNames.data_has_header_line_step] == "no":
            print("Reinterpreting your first row as a record, not headers.")
            self.state["filelist_data_list"].insert(
                0, self.state["filelist_data_list"][0]
            )


class HasSpeakerStep(Step):
    DEFAULT_NAME = StepNames.data_has_speaker_value_step
    choices = ("no", "yes")

    def prompt(self):
        if self.state[StepNames.filelist_format_step] == "festival":
            return "no"
        elif len(self.state.get("selected_headers", [])) >= len(
            self.state["filelist_data_list"][0]
        ):
            print("No columns left, we will assume you have no speaker column.")
            return "no"
        else:
            return get_response_from_menu_prompt(
                prompt_text="Does your data have a column/value for the speaker?",
                choices=self.choices,
            )

    def validate(self, response):
        return response in self.choices

    def effect(self):
        if self.state[StepNames.data_has_speaker_value_step] == "yes":
            self.tour.add_step(
                HeaderStep(
                    name=StepNames.speaker_header_step,
                    prompt_text="These are the remaining values from the first row in your data. Which column contains the [bold blue]speaker?",
                    header_name="speaker",
                    state_subset=self.state_subset,
                ),
                self,
            )


class HasLanguageStep(Step):
    DEFAULT_NAME = StepNames.data_has_language_value_step
    choices = ("no", "yes")

    def prompt(self):
        if self.state[StepNames.filelist_format_step] == "festival":
            return "no"
        elif len(self.state.get("selected_headers", [])) >= len(
            self.state["filelist_data_list"][0]
        ):
            print("No columns left, we will assume you have no language column.")
            return "no"
        else:
            return get_response_from_menu_prompt(
                prompt_text="Does your data have a column/value for the language?",
                choices=self.choices,
            )

    def validate(self, response):
        return response in self.choices

    def effect(self):
        if self.state[StepNames.data_has_language_value_step] == "yes":
            self.tour.add_step(
                LanguageHeaderStep(
                    name=StepNames.language_header_step,
                    prompt_text="These are the remaining values from the first row in your data. Which column contains the [bold blue]language?",
                    header_name="language",
                    state_subset=self.state_subset,
                ),
                self,
            )

        else:
            self.tour.add_step(SelectLanguageStep(state_subset=self.state_subset), self)


class SelectLanguageStep(Step):
    DEFAULT_NAME = StepNames.select_language_step

    def prompt(self):
        from g2p import get_arpabet_langs

        from everyvoice.text.phonemizer import AVAILABLE_G2P_ENGINES

        g2p_langs_full = get_arpabet_langs()[1]
        print(
            "Note: if your dataset has more than one language in it, you will have to provide a 'language' column to indicate the language of each sample, because the configuration wizard can't guess!"
        )
        # TODO: currently we only support the languages from g2p, but we should add more
        supported_langs = list(AVAILABLE_G2P_ENGINES)
        supported_langs_choices = ["[und]: my language isn't here"] + [
            f"[{k}]: {g2p_langs_full.get(k, 'Unknown')}" for k in supported_langs
        ]
        return get_response_from_menu_prompt(
            choices=supported_langs_choices,
            title="Which of the following supported languages is the language of your dataset?",
            multi=False,
            search=True,
        )

    def validate(self, response):
        return isinstance(response, str)

    def effect(self):
        # Rename unselected headers to unknown:
        self.state["filelist_headers"] = rename_unknown_headers(
            self.state["filelist_headers"]
        )
        # re-parse data:
        reload_filelist_data_as_dict(self.state)
        # Apply the language code:
        isocode = get_iso_code(self.response)
        # Apply text conversions and get target training representation
        self.state["model_target_training_text_representation"] = (
            apply_automatic_text_conversions(
                self.state["filelist_data"],
                self.state[StepNames.filelist_text_representation_step],
                global_isocode=isocode,
            )
        )


def reload_filelist_data_as_dict(state):
    """Given a tour or step's state, reload the filelist_data as a dict if
    that was not already done."""
    data = state.get("filelist_data", None)
    if data and isinstance(data[0], dict):
        # data is already a dict, no need to do anything
        return

    # reparse data
    filelist_path = state.get(StepNames.filelist_step)
    if not isinstance(filelist_path, Path):
        filelist_path = Path(filelist_path).expanduser()
    if state.get(StepNames.filelist_format_step, None) in [
        "psv",
        "tsv",
        "csv",
    ]:
        state["filelist_data"] = generic_xsv_filelist_reader(
            filelist_path,
            delimiter=state.get("filelist_delimiter"),
            fieldnames=state["filelist_headers"],
            file_has_header_line=(
                state.get(StepNames.data_has_header_line_step, "yes") == "yes"
            ),
        )
    else:
        state["filelist_data"] = read_festival(
            filelist_path,
            text_field_name=state.get(
                StepNames.filelist_text_representation_step, "text"
            ),
        )


def get_iso_code(language):
    if language is None:
        return None
    result = re.search(r"\[[\w-]*\]", language)
    if result is None:
        return language
    else:
        return result.group()[1:-1]


class TextProcessingStep(Step):
    DEFAULT_NAME = StepNames.text_processing_step
    process_lookup = {
        0: {"fn": lower, "desc": "lowercase"},
        1: {"fn": nfc_normalize, "desc": "NFC Normalization"},
    }

    def prompt(self):
        return get_response_from_menu_prompt(
            prompt_text=f"Which of the following text transformations would like to apply to your dataset's {self.state[StepNames.filelist_text_representation_step]}?",
            choices=(
                "Lowercase",
                "NFC Normalization - See here for more information: https://withblue.ink/2019/03/11/why-you-need-to-normalize-unicode-strings.html",
            ),
            multi=True,
            search=False,
            return_indices=True,
        )

    def validate(self, response):
        return True

    def effect(self):
        # Apply the selected text processing processes
        if "symbols" not in self.state:
            self.state["symbols"] = {}
        if self.response:
            text_index = self.state["filelist_headers"].index(
                self.state[StepNames.filelist_text_representation_step]
            )
            for process in self.response:
                process_fn = self.process_lookup[process]["fn"]
                for i in tqdm(
                    range(len(self.state["filelist_data_list"])),
                    desc=f"Applying {self.process_lookup[process]['desc']} to data",
                ):
                    self.state["filelist_data_list"][i][text_index] = process_fn(
                        self.state["filelist_data_list"][i][text_index]
                    )


class SoxEffectsStep(Step):
    DEFAULT_NAME = StepNames.sox_effects_step

    def prompt(self):
        return get_response_from_menu_prompt(
            prompt_text="Which of the following audio preprocessing options would you like to apply?",
            choices=(
                "Normalization (-3.0dB)",
                "Remove Silence at start and end",
                "Remove Silence throughout",
            ),
            multi=True,
            search=False,
            return_indices=True,
        )

    def validate(self, response):
        return True

    def effect(self):
        audio_effects = {
            0: [["norm", "-3.0"]],
            1: [
                [
                    "silence",
                    "1",
                    "0.1",
                    "0.1%",
                ],
                ["reverse"],  # reverse the clip to trim silence from end
                ["silence", "1", "0.1", "0.1%"],
                [
                    "reverse"
                ],  # reverse the clip again to revert to the right direction :)
            ],
            2: [["silence", "1", "0.1", "1.0%", "-1", "0.4", "1%"]],
        }
        self.state["sox_effects"] = [["channels", "1"]]
        if self.response:
            for effect in self.response:
                self.state["sox_effects"] += audio_effects[effect]


class SymbolSetStep(Step):
    DEFAULT_NAME = StepNames.symbol_set_step

    def prompt(self):
        # TODO: This is a bit of a weird step, since it doesn't really prompt anything, it just applies the effect of trying to find
        #       character graphemes/phones. I'd still like to keep it here, since we might add more to this step in the future, and
        #       I don't want to lump the grapheme clustering logic into the effect of another step.
        print(
            f"We will now read your entire dataset and try to determine the characters and/or phones in your dataset according to Unicode Grapheme clustering rules. Please carefully check your {TEXT_CONFIG_FILENAME_PREFIX}.yaml file (which is created at the end of the wizard) and adjust the symbol set as appropriate. If your language uses standard punctuation symbols to represent sounds, it is extra important that you go remove any of these symbols from the punctuation categories."
        )
        return True

    def validate(self, response):
        return bool(response)

    def effect(self):
        character_graphemes = set()
        phone_graphemes = set()
        for item in tqdm(
            self.state["filelist_data"], desc="Finding all symbols in your dataset"
        ):
            if "characters" in item:
                character_graphemes.update(guess_graphemes_in_text(item["characters"]))
            if "phones" in item:
                phone_graphemes.update(guess_ipa_phones_in_text(item["phones"]))
        if not phone_graphemes and not character_graphemes:
            return
        symbols = {}
        if character_graphemes:
            symbols["characters"] = sorted(list(character_graphemes))
        if phone_graphemes:
            symbols["phones"] = sorted(list(phone_graphemes))
        self.state[StepNames.symbol_set_step] = symbols


def get_dataset_steps(dataset_index=0):
    return [
        FilelistStep(state_subset=f"dataset_{dataset_index}"),
        [
            DatasetPermissionStep(state_subset=f"dataset_{dataset_index}"),
            FilelistFormatStep(state_subset=f"dataset_{dataset_index}"),
            FilelistTextRepresentationStep(state_subset=f"dataset_{dataset_index}"),
            TextProcessingStep(state_subset=f"dataset_{dataset_index}"),
            HasSpeakerStep(state_subset=f"dataset_{dataset_index}"),
            HasLanguageStep(state_subset=f"dataset_{dataset_index}"),
            WavsDirStep(state_subset=f"dataset_{dataset_index}"),
            ValidateWavsStep(state_subset=f"dataset_{dataset_index}"),
            SymbolSetStep(state_subset=f"dataset_{dataset_index}"),
            SoxEffectsStep(state_subset=f"dataset_{dataset_index}"),
            DatasetNameStep(state_subset=f"dataset_{dataset_index}"),
        ],
    ]


if __name__ == "__main__":
    tour = Tour(
        name="Dataset Tour",
        # steps = [TextProcessingStep(name='test')]
        steps=get_dataset_steps(),
    )
    tour.visualize()
    tour.run()



import json

import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
import torchaudio
from loguru import logger

from everyvoice.model.e2e.config import EveryVoiceConfig
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.model import (
    FastSpeech2,
)
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.type_definitions_heavy import (
    Stats,
)
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.utils.heavy import (
    plot_mel,
)
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.model import HiFiGAN
from everyvoice.text.lookups import lookuptables_from_config
from everyvoice.utils.heavy import dynamic_range_compression_torch, expand


class EveryVoice(pl.LightningModule):
    def __init__(self, config: EveryVoiceConfig):
        super().__init__()
        # Required by PyTorch Lightning for our manual optimization
        self.automatic_optimization = False
        self.config = config
        self.batch_size = config.training.batch_size
        if self.config.training.feature_prediction_checkpoint is not None:
            logger.info(
                f"Loading FastSpeech2 model from checkpoint {self.config.training.feature_prediction_checkpoint}"
            )
            self.feature_prediction: FastSpeech2 = FastSpeech2.load_from_checkpoint(
                self.config.training.feature_prediction_checkpoint
            )
            self.feature_prediction.config = self.config.feature_prediction
        else:
            lang2id, speaker2id = lookuptables_from_config(config.feature_prediction)
            # TODO: What about when we are fine-tuning? Do the bins in the Variance Adaptor not change? https://github.com/roedoejet/FastSpeech2_lightning/issues/28
            with open(
                config.feature_prediction.preprocessing.save_dir / "stats.json"
            ) as f:
                stats: Stats = Stats(**json.load(f))
            self.feature_prediction = FastSpeech2(
                config.feature_prediction,
                stats=stats,
                lang2id=lang2id,
                speaker2id=speaker2id,
            )
        if self.config.training.vocoder_checkpoint is not None:
            logger.info(
                f"Loading HiFiGAN model from checkpoint {self.config.training.vocoder_checkpoint}"
            )
            self.vocoder: HiFiGAN = HiFiGAN.load_from_checkpoint(
                self.config.training.vocoder_checkpoint
            )
            self.vocoder.config = self.config.vocoder
        else:
            self.vocoder = HiFiGAN(config.vocoder)
        # Required by PyTorch Lightning for our manual optimization
        self.vocoder.automatic_optimization = False
        self.feature_prediction.automatic_optimization = False
        self.sampling_rate_change = (
            self.config.vocoder.preprocessing.audio.output_sampling_rate
            // self.config.vocoder.preprocessing.audio.input_sampling_rate
        )
        self.input_frame_segment_size = (
            self.config.vocoder.preprocessing.audio.vocoder_segment_size
            // self.config.vocoder.preprocessing.audio.fft_hop_size
        )
        self.use_segments = True

    def forward(self, batch, inference=False):
        feature_prediction = self.feature_prediction.forward(batch)
        vocoder_x = feature_prediction[self.feature_prediction.output_key]
        if not inference and self.use_segments:
            vocoder_x = torch.stack(
                [
                    x[
                        batch["segment_start_frame"][i] : batch["segment_start_frame"][
                            i
                        ]
                        + self.input_frame_segment_size
                    ]
                    for i, x in enumerate(vocoder_x)
                ]
            )
        # Generate Waveform
        if self.config.vocoder.model.istft_layer:
            mag, phase = self.vocoder(vocoder_x.transpose(1, 2))
            vocoder_prediction = self.vocoder.inverse_spectral_transform(
                mag * torch.exp(phase * 1j)
            ).unsqueeze(-2)
        else:
            vocoder_prediction = self.vocoder(vocoder_x.transpose(1, 2))
        return feature_prediction, vocoder_prediction

    def vocoder_loss(self, y, y_mel, y_hat, y_hat_mel, mode="training"):
        _, y_df_hat_g, fmap_f_r, fmap_f_g = self.vocoder.mpd(y, y_hat)
        _, y_ds_hat_g, fmap_s_r, fmap_s_g = self.vocoder.msd(y, y_hat)
        loss_fm_f = self.vocoder.feature_loss(fmap_f_r, fmap_f_g)
        loss_fm_s = self.vocoder.feature_loss(fmap_s_r, fmap_s_g)
        loss_gen_f, _ = self.vocoder.generator_loss(y_df_hat_g)
        loss_gen_s, _ = self.vocoder.generator_loss(y_ds_hat_g)
        loss_mel = torch.nn.functional.l1_loss(y_mel, y_hat_mel) * 45
        gen_loss_total = 0
        # Log Losses
        self.log(f"{mode}/vocoder/gen/loss_fmap_f", loss_fm_f, prog_bar=False)
        self.log(f"{mode}/vocoder/gen/loss_fmap_s", loss_fm_s, prog_bar=False)
        self.log(f"{mode}/vocoder/gen/loss_gen_f", loss_gen_f, prog_bar=False)
        self.log(f"{mode}/vocoder/gen/loss_gen_s", loss_gen_s, prog_bar=False)
        self.log(f"{mode}/vocoder/gen/mel_spec_error", loss_mel / 45, prog_bar=False)
        gen_loss_total = loss_gen_s + loss_gen_f + loss_fm_s + loss_fm_f + loss_mel
        self.log(f"{mode}/vocoder/gen/gen_loss_total", gen_loss_total, prog_bar=False)
        return gen_loss_total

    def training_step(self, batch, batch_idx):
        y = batch["audio"]
        y_mel = batch["audio_mel"]
        # Train Feature Prediction & Vocoder Generator
        optim_fp, optim_g, optim_d = self.optimizers()
        scheduler_fp, scheduler_g, scheduler_d = self.lr_schedulers()
        # Generate Predictions
        fp_output, generated_wav = self(batch)
        # Create Mel of generated wav
        generated_mel_spec = dynamic_range_compression_torch(
            self.vocoder.spectral_transform(generated_wav).squeeze(1)[:, :, 1:]
        )

        # Train Vocoder Discriminator
        if self.global_step >= self.vocoder.config.training.generator_warmup_steps:
            optim_d.zero_grad()
            # MPD
            y_df_hat_r, y_df_hat_g, _, _ = self.vocoder.mpd(y, generated_wav.detach())
            loss_disc_f, _, _ = self.vocoder.discriminator_loss(y_df_hat_r, y_df_hat_g)
            self.log("training/disc/mpd_loss", loss_disc_f, prog_bar=False)
            # MSD
            y_ds_hat_r, y_ds_hat_g, _, _ = self.vocoder.msd(y, generated_wav.detach())
            loss_disc_s, _, _ = self.vocoder.discriminator_loss(y_ds_hat_r, y_ds_hat_g)
            self.log("training/disc/msd_loss", loss_disc_s, prog_bar=False)
            # WGAN
            if self.vocoder.use_wgan:
                for p in self.vocoder.msd.parameters():
                    p.data.clamp_(
                        -self.config.training.wgan_clip_value,
                        self.config.training.wgan_clip_value,
                    )
                for p in self.vocoder.mpd.parameters():
                    p.data.clamp_(
                        -self.config.training.wgan_clip_value,
                        self.config.training.wgan_clip_value,
                    )
            # calculate loss
            disc_loss_total = loss_disc_s + loss_disc_f
            # manual optimization because Pytorch Lightning 2.0+ doesn't handle automatic optimization for multiple optimizers
            # use .backward for now, but maybe switch to self.manual_backward() in the future: https://github.com/Lightning-AI/lightning/issues/18740
            disc_loss_total.backward(retain_graph=True)
            # clip gradients
            self.clip_gradients(
                optim_d, gradient_clip_val=1.0, gradient_clip_algorithm="norm"
            )
            optim_d.step()
            # step in the scheduler every epoch
            if self.trainer.is_last_batch:
                scheduler_d.step()
            # log discriminator loss
            self.log("training/disc/d_loss_total", disc_loss_total, prog_bar=False)
        # train generator
        optim_fp.zero_grad()
        optim_g.zero_grad()
        # Calculate Feature Prediction Loss
        fp_losses: torch.Tensor = self.feature_prediction.loss(
            fp_output, batch, self.current_epoch
        )
        self.log_dict(
            {
                f"training/feature_prediction/{k}_loss": v.item()
                for k, v in fp_losses.items()
            }
        )

        # Calculate Generator Loss
        loss_mel = F.l1_loss(y_mel, generated_mel_spec) * 45
        if self.global_step >= self.vocoder.config.training.generator_warmup_steps:
            _, y_df_hat_g, fmap_f_r, fmap_f_g = self.vocoder.mpd(y, generated_wav)
            _, y_ds_hat_g, fmap_s_r, fmap_s_g = self.vocoder.msd(y, generated_wav)
            loss_fm_f = self.vocoder.feature_loss(fmap_f_r, fmap_f_g)
            loss_fm_s = self.vocoder.feature_loss(fmap_s_r, fmap_s_g)
            # loss_gen_f = -torch.mean(y_df_hat_g)
            # loss_gen_s = -torch.mean(y_ds_hat_g)
            loss_gen_f, _ = self.vocoder.generator_loss(y_df_hat_g)
            loss_gen_s, _ = self.vocoder.generator_loss(y_ds_hat_g)
            self.log("training/gen/loss_fmap_f", loss_fm_f, prog_bar=False)
            self.log("training/gen/loss_fmap_s", loss_fm_s, prog_bar=False)
            self.log("training/gen/loss_gen_f", loss_gen_f, prog_bar=False)
            self.log("training/gen/loss_gen_s", loss_gen_s, prog_bar=False)
            gen_loss_total = loss_gen_s + loss_gen_f + loss_fm_s + loss_fm_f + loss_mel
        else:
            gen_loss_total = loss_mel
        # manual optimization because Pytorch Lightning 2.0+ doesn't handle automatic optimization for multiple optimizers
        # use .backward for now, but maybe switch to self.manual_backward() in the future: https://github.com/Lightning-AI/lightning/issues/18740
        # self.manual_backward(gen_loss_total)
        gen_loss_total.backward(retain_graph=True)
        # clip gradients
        self.clip_gradients(
            optim_d, gradient_clip_val=1.0, gradient_clip_algorithm="norm"
        )
        # fp_losses["total"].backward()
        optim_g.step()
        optim_fp.step()
        # step in the scheduler every epoch
        if self.trainer.is_last_batch:
            scheduler_g.step()
            scheduler_fp.step()
            # log generator loss
        self.log("training/gen/gen_loss_total", gen_loss_total, prog_bar=False)
        self.log("training/gen/mel_spec_error", loss_mel / 45, prog_bar=False)

    def validation_step(self, batch, batch_idx):
        fp_output, vocoder_y_hat = self(batch)
        fp_losses = self.feature_prediction.loss(fp_output, batch, self.current_epoch)
        generated_mel_spec = dynamic_range_compression_torch(
            self.vocoder.spectral_transform(vocoder_y_hat).squeeze(1)[:, :, 1:]
        )
        self.log_dict(
            {
                f"validation/feature_prediction/{k}_loss": v.item()
                for k, v in fp_losses.items()
            }
        )
        val_err_tot = torch.nn.functional.l1_loss(
            batch["audio_mel"], generated_mel_spec
        ).item()
        self.log("validation/mel_spec_error", val_err_tot, prog_bar=False)
        if self.global_step == 0:
            audio, _ = torchaudio.load(
                self.config.feature_prediction.preprocessing.save_dir
                / "audio"
                / "--".join(
                    [
                        batch["basename"][0],
                        batch["speaker"][0],
                        batch["language"][0],
                        f"audio-{self.config.feature_prediction.preprocessing.audio.input_sampling_rate}.wav",
                    ]
                )
            )
            # Log ground truth audio
            self.logger.experiment.add_audio(
                f"gt/wav_{batch['basename'][0]}",
                audio,
                self.global_step,
                self.config.feature_prediction.preprocessing.audio.output_sampling_rate,
            )
        if batch_idx == 0:
            fp_output, synthesis_wav_output = self(batch, inference=True)
            duration_np = fp_output["duration_target"][0].cpu().numpy()
            self.logger.experiment.add_figure(
                f"pred/spec_{batch['basename'][0]}",
                plot_mel(
                    [
                        {
                            "mel": np.swapaxes(batch["mel"][0].cpu().numpy(), 0, 1),
                            "pitch": expand(
                                batch["pitch"][0].cpu().numpy(), duration_np
                            ),
                            "energy": expand(
                                batch["energy"][0].cpu().numpy(), duration_np
                            ),
                        },
                        {
                            "mel": np.swapaxes(
                                fp_output[self.feature_prediction.output_key][0]
                                .cpu()
                                .numpy(),
                                0,
                                1,
                            ),
                            "pitch": expand(
                                fp_output["pitch_prediction"][0].cpu().numpy(),
                                duration_np,
                            ),
                            "energy": expand(
                                fp_output["energy_prediction"][0].cpu().numpy(),
                                duration_np,
                            ),
                        },
                    ],
                    self.feature_prediction.stats,
                    ["Ground-Truth Spectrogram", "Synthesized Spectrogram"],
                ),
                self.global_step,
            )
            self.logger.experiment.add_audio(
                f"pred/wav_{batch['basename'][0]}",
                synthesis_wav_output,
                self.global_step,
                self.config.vocoder.preprocessing.audio.output_sampling_rate,
            )
            if self.config.vocoder.model.istft_layer:
                mag, phase = self.vocoder(batch["mel"].transpose(1, 2))
                copy_synthesis = self.vocoder.inverse_spectral_transform(
                    mag * torch.exp(phase * 1j)
                ).unsqueeze(-2)
            else:
                copy_synthesis = self.vocoder(batch["mel"].transpose(1, 2))
            self.logger.experiment.add_audio(
                f"copy-synthesis/wav_{batch['basename'][0]}",
                copy_synthesis,
                self.global_step,
                self.config.vocoder.preprocessing.audio.output_sampling_rate,
            )
        return fp_losses["total"] + val_err_tot

    def configure_optimizers(self):
        fp_optims, fp_schedulers = self.feature_prediction.configure_optimizers()
        vocoder_optims, vocoder_schedulers = self.vocoder.configure_optimizers()
        return fp_optims + vocoder_optims, fp_schedulers + vocoder_schedulers



from pathlib import Path
from typing import Any, Callable, Optional, Sequence

from loguru import logger
from pydantic import PlainSerializer, WithJsonSchema
from pydantic.functional_validators import BeforeValidator
from typing_extensions import Annotated

from everyvoice.utils import load_config_from_json_or_yaml_path

from .validation_helpers import (
    callable_to_string,
    directory_path_must_exist,
    path_is_a_directory,
    relative_to_absolute_path,
    string_to_callable,
)


def load_partials(
    pre_validated_model_dict: dict[Any, Any],
    partial_keys: Sequence[str],
    config_path: Optional[Path] = None,
):
    """Loads all partials based on a list of partial keys. For this to work,
    your model must have a {key}_config_file: Optional[FilePath] = None field
    defined, and you must have a model_validator(mode="before") that runs this
    function.
    """
    # If there's nothing there, just return the dict
    if not pre_validated_model_dict:
        return pre_validated_model_dict
    # Otherwise go through each key and load it in
    for key in partial_keys:
        key_for_path_to_partial = f"path_to_{key}_config_file"  # I added path_to_ because model_* is a restricted namespace on pydantic models
        if (
            key_for_path_to_partial in pre_validated_model_dict
            and pre_validated_model_dict[key_for_path_to_partial]
        ):
            subconfig_path = Path(pre_validated_model_dict[key_for_path_to_partial])
            if not subconfig_path.is_absolute() and config_path is not None:
                subconfig_path = (config_path.parent / subconfig_path).resolve()
            pre_validated_model_dict[key_for_path_to_partial] = subconfig_path
            # anything defined in the key will override the path
            # so audio would override any values in path_to_audio_config_file
            if key in pre_validated_model_dict:
                logger.info(
                    f"You have both the key {key} and {key_for_path_to_partial} defined in your configuration. We will override values from {key_for_path_to_partial} with values from {key}"
                )
                if isinstance(pre_validated_model_dict[key], dict):
                    pre_validated_model_dict[key] = {
                        **load_config_from_json_or_yaml_path(subconfig_path),
                        **pre_validated_model_dict[key],
                    }
                else:
                    try:
                        # Maybe a model was passed
                        pre_validated_model_dict[key] = {
                            **load_config_from_json_or_yaml_path(subconfig_path),
                            **pre_validated_model_dict[key].model_dump(),
                        }
                    except AttributeError:
                        # If model_dump() doesn't exist then don't try and merge anything just continue and let the model raise a pydantic.ValidationError
                        pass
            else:
                pre_validated_model_dict[key] = load_config_from_json_or_yaml_path(
                    subconfig_path
                )
    return pre_validated_model_dict


PossiblySerializedCallable = Annotated[
    Callable,
    BeforeValidator(string_to_callable),
    PlainSerializer(callable_to_string, return_type=str),
    WithJsonSchema({"type": "string"}, mode="serialization"),  # noqa: F821
    WithJsonSchema({"type": "string"}, mode="validation"),  # noqa: F821
]

PossiblyRelativePath = Annotated[Path, BeforeValidator(relative_to_absolute_path)]

# [Ordering of validators within Annotated](https://docs.pydantic.dev/latest/concepts/validators/#ordering-of-validators-within-annotated)
# Order of validation metadata within Annotated matters. Validation goes from
# right to left and back. That is, it goes from right to left running all
# "before" validators (or calling into "wrap" validators), then left to right
# back out calling all "after" validators.
PossiblyRelativePathMustExist = Annotated[
    Path,
    BeforeValidator(path_is_a_directory),
    BeforeValidator(directory_path_must_exist),
    BeforeValidator(relative_to_absolute_path),
]



import math
from enum import Enum
from pathlib import Path
from typing import Any, Optional

from everyvoice.config.preprocessing_config import PreprocessingConfig
from everyvoice.config.shared_types import (
    AdamOptimizer,
    AdamWOptimizer,
    BaseModelWithContact,
    BaseTrainingConfig,
    ConfigModel,
    RMSOptimizer,
    init_context,
)
from everyvoice.config.utils import PossiblySerializedCallable, load_partials
from everyvoice.utils import (
    load_config_from_json_or_yaml_path,
    original_hifigan_leaky_relu,
)
from pydantic import (
    Field,
    FilePath,
    ValidationInfo,
    field_serializer,
    field_validator,
    model_validator,
)


# NOTE: We need to derive from both str and Enum if we want `HiFiGANResblock.one == "one"` to be True.
#    Otherwise, `HiFiGANResblock.one == "one"` will be false.
#    [Python Enum Comparisons](https://docs.python.org/3/howto/enum.html#comparisons)
#    Comparisons against non-enumeration values will always compare not equal.
#    In [1]: from enum import Enum
#    In [2]: class A(Enum):
#       ...:     a = "a"
#       ...:     b = "b"
#       ...:
#    In [3]: A.a == "a"
#    Out[3]: False
#    In [4]: class S(str, Enum):
#       ...:     a = "a"
#       ...:     b = "b"
#       ...:
#    In [5]: S.a == "a"
#    Out[5]: True
# NOTE: The reason behind using an enum is that we actually want to compare to
#    an enum "value" and not to its string representation like so:
#    In [5]: a = A.a
#    In [6]: a is A.a
#    Out[6]: True
class HiFiGANResblock(str, Enum):
    one = "1"
    two = "2"


class HiFiGANTrainTypes(str, Enum):
    original = "original"
    wgan = "wgan"
    # wgan_gp = "wgan-gp"


class HiFiGANModelConfig(ConfigModel):
    resblock: HiFiGANResblock = Field(
        HiFiGANResblock.one,
        description="Which resblock to use. See Kong et. al. 2020: https://arxiv.org/abs/2010.05646",
    )
    upsample_rates: list[int] = Field(
        [8, 8, 2, 2],
        description="The stride of each convolutional layer in the upsampling module.",
    )
    upsample_kernel_sizes: list[int] = Field(
        [16, 16, 4, 4],
        description="The kernel size of each convolutional layer in the upsampling module.",
    )
    upsample_initial_channel: int = Field(
        512,
        description="The number of dimensions to project the Mel inputs to before being passed to the resblock.",
    )
    resblock_kernel_sizes: list[int] = Field(
        [3, 7, 11],
        description="The kernel size of each convolutional layer in the resblock.",
    )
    resblock_dilation_sizes: list[list[int]] = Field(
        [[1, 3, 5], [1, 3, 5], [1, 3, 5]],
        description="The dilations of each convolution in each layer of the resblock.",
    )
    activation_function: PossiblySerializedCallable = Field(
        original_hifigan_leaky_relu, description="The activation function to use."
    )
    istft_layer: bool = Field(
        False,
        description="Whether to predict phase and magnitude values and use an inverse Short-Time Fourier Transform instead of predicting a waveform directly. See Kaneko et. al. 2022: https://arxiv.org/abs/2203.02395",
    )
    msd_layers: int = Field(
        3, description="The number of layers to use in the Multi-Scale Discriminator."
    )
    mpd_layers: list[int] = Field(
        [2, 3, 5, 7, 11],
        description="The size of each layer in the Multi-Period Discriminator.",
    )

    @field_serializer("resblock")
    def convert_enum(self, resblock: HiFiGANResblock):
        return resblock.value

    @field_validator("resblock", mode="after")
    @classmethod
    def convert_to_HiFiGANResblock(
        cls,
        v: str | HiFiGANResblock,
        _info: ValidationInfo,
    ) -> HiFiGANResblock:
        return HiFiGANResblock(v)


class HiFiGANTrainingConfig(BaseTrainingConfig):
    generator_warmup_steps: int = Field(
        0,
        description="The number of steps to run through before activating the discriminators.",
    )
    gan_type: HiFiGANTrainTypes = Field(
        HiFiGANTrainTypes.original,
        description="The type of GAN to use. Can be set to either 'original' for a vanilla GAN, or 'wgan' for a Wasserstein GAN that clips gradients.",
    )
    optimizer: AdamOptimizer | AdamWOptimizer | RMSOptimizer = Field(
        default_factory=AdamWOptimizer,
        description="Configuration settings for the optimizer.",
    )
    wgan_clip_value: float = Field(
        0.01, description="The gradient clip value when gan_type='wgan'."
    )
    use_weighted_sampler: bool = Field(
        False,
        description="Whether to use a sampler which oversamples from the minority language or speaker class for balanced training.",
    )
    finetune: bool = Field(
        False,
        description="Whether to read spectrograms from 'preprocessed/synthesized_spec' instead of 'preprocessed/spec'. This is used when finetuning a pretrained spec-to-wav (vocoder) model using the outputs of a trained text-to-spec (feature prediction network) model.",
    )

    @field_serializer("gan_type")
    def convert_enum(self, gan_type: HiFiGANTrainTypes):
        return gan_type.value

    @field_validator("gan_type", mode="after")
    @classmethod
    def convert_to_HiFiGANTrainTypes(
        cls,
        v: str | HiFiGANTrainTypes,
        _info: ValidationInfo,
    ) -> HiFiGANTrainTypes:
        return HiFiGANTrainTypes(v)


class HiFiGANConfig(BaseModelWithContact):
    model: HiFiGANModelConfig = Field(
        default_factory=HiFiGANModelConfig,
        description="The model configuration settings.",
    )
    path_to_model_config_file: Optional[FilePath] = Field(
        None, description="The path of a model configuration file."
    )
    training: HiFiGANTrainingConfig = Field(
        default_factory=HiFiGANTrainingConfig,
        description="The training configuration hyperparameters.",
    )
    path_to_training_config_file: Optional[FilePath] = Field(
        None, description="The path of a training configuration file."
    )
    preprocessing: PreprocessingConfig = Field(
        default_factory=PreprocessingConfig,
        description="The preprocessing configuration, including information about audio settings.",
    )
    path_to_preprocessing_config_file: Optional[FilePath] = Field(
        None, description="The path of a preprocessing configuration file."
    )

    @model_validator(mode="before")  # type: ignore
    def load_partials(self: dict[Any, Any], info: ValidationInfo):
        config_path = (
            info.context.get("config_path", None) if info.context is not None else None
        )
        return load_partials(
            self,
            ("model", "training", "preprocessing"),
            config_path=config_path,
        )

    @staticmethod
    def load_config_from_path(path: Path) -> "HiFiGANConfig":
        """Load a config from a path"""
        config = load_config_from_json_or_yaml_path(path)
        with init_context({"config_path": path}):
            config = HiFiGANConfig(**config)
        return config

    @model_validator(mode="after")
    def check_upsample_rate_consistency(self) -> "HiFiGANConfig":
        # helper variables
        preprocessing_config: PreprocessingConfig = self.preprocessing
        model_config: HiFiGANModelConfig = self.model
        sampling_rate = preprocessing_config.audio.input_sampling_rate
        upsampled_sampling_rate = preprocessing_config.audio.output_sampling_rate
        upsample_rate = upsampled_sampling_rate // sampling_rate
        upsampled_hop_size = upsample_rate * preprocessing_config.audio.fft_hop_size
        upsample_rate_product = math.prod(model_config.upsample_rates)
        # check that same number of kernels and kernel sizes exist
        if len(model_config.upsample_kernel_sizes) != len(model_config.upsample_rates):
            raise ValueError(
                "Number of upsample kernel sizes must match number of upsample rates"
            )
        # Check that kernel sizes are not less than upsample rates and are evenly divisible
        for kernel_size, upsample_rate in zip(
            model_config.upsample_kernel_sizes, model_config.upsample_rates
        ):
            if kernel_size < upsample_rate:
                raise ValueError(
                    f"Upsample kernel size: {kernel_size} must be greater than upsample rate: {upsample_rate}"
                )
            if kernel_size % upsample_rate != 0:
                raise ValueError(
                    f"Upsample kernel size: {kernel_size} must be evenly divisible by upsample rate: {upsample_rate}"
                )
        # check that upsample rate is even multiple of target sampling rate
        if upsampled_sampling_rate % sampling_rate != 0:
            raise ValueError(
                f"Target sampling rate: {upsampled_sampling_rate} must be an even multiple of input sampling rate: {sampling_rate}"
            )
        # check that the upsampling hop size is equal to product of upsample rates
        if model_config.istft_layer:
            upsampled_hop_size //= 4  # istft upsamples the rest
        # check that upsampled hop size is equal to product of upsampling rates
        if upsampled_hop_size != upsample_rate_product:
            raise ValueError(
                f"Upsampled hop size: {upsampled_hop_size} must be equal to product of upsample rates: {upsample_rate_product}"
            )
        # check that the segment size is divisible by product of upsample rates
        if preprocessing_config.audio.vocoder_segment_size % upsample_rate_product != 0:
            raise ValueError(
                f"Vocoder segment size: {preprocessing_config.audio.vocoder_segment_size} must be divisible by product of upsample rates: {upsample_rate_product}"
            )

        return self



import os
from pathlib import Path
from typing import Callable, Optional, Union

import pytorch_lightning as pl
import torch
from torch.utils.data import DataLoader

from everyvoice.dataloader.imbalanced_sampler import ImbalancedDatasetSampler
from everyvoice.model.aligner.config import AlignerConfig
from everyvoice.model.e2e.config import EveryVoiceConfig
from everyvoice.model.feature_prediction.config import FeaturePredictionConfig
from everyvoice.model.vocoder.config import VocoderConfig


class BaseDataModule(pl.LightningDataModule):
    def __init__(
        self,
        config: Union[
            AlignerConfig, VocoderConfig, FeaturePredictionConfig, EveryVoiceConfig
        ],
        inference_output_dir: Optional[Path] = None,
    ):
        super().__init__()
        self.collate_fn: Union[Callable, None] = None
        self.config = config
        self.use_weighted_sampler = False
        self.inference_output_dir = inference_output_dir
        if self.inference_output_dir is not None:
            self.inference_output_dir.mkdir(exist_ok=True, parents=True)
            self.predict_path = self.inference_output_dir / "latest_predict_data.pth"
        self.train_path = os.path.join(
            self.config.training.logger.save_dir,
            self.config.training.logger.name,
            "train_data.pth",
        )
        self.val_path = os.path.join(
            self.config.training.logger.save_dir,
            self.config.training.logger.name,
            "val_data.pth",
        )

    def setup(self, stage: Optional[str] = None):
        # load it back here
        if stage == "fit":
            self.train_dataset = torch.load(self.train_path)
            self.val_dataset = torch.load(self.val_path)
        if stage == "predict":
            self.predict_dataset = torch.load(self.predict_path)

    def train_dataloader(self):
        sampler = (
            ImbalancedDatasetSampler(self.train_dataset)
            if self.use_weighted_sampler
            else None
        )
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            num_workers=self.config.training.train_data_workers,
            pin_memory=False,
            drop_last=True,
            collate_fn=self.collate_fn,
            sampler=sampler,
        )

    def predict_dataloader(self):
        return DataLoader(
            self.predict_dataset,
            batch_size=self.batch_size,
            num_workers=self.config.training.train_data_workers,
            pin_memory=False,
            drop_last=False,
            collate_fn=self.collate_fn,
        )

    def val_dataloader(self):
        sampler = (
            ImbalancedDatasetSampler(self.val_dataset)
            if self.use_weighted_sampler
            else None
        )
        return DataLoader(
            self.val_dataset,
            batch_size=1,
            num_workers=0,
            pin_memory=False,
            drop_last=True,
            collate_fn=self.collate_fn,
            sampler=sampler,
        )

    def prepare_data(self):
        # Override this method
        raise NotImplementedError(
            "This method should be implemented by the child class"
        )

    def load_dataset(self):
        # Override this method
        raise NotImplementedError(
            "The base data module does not have a method implemented for loading a dataset. Please use another Data Loader that inherits the BaseDataModule class."
        )



""" EveryVoice performs grapheme-to-phoneme conversion based on language IDs
    All g2p engines must return tokenized characters.
"""

import re
from typing import Callable
from unicodedata import normalize

from g2p import get_arpabet_langs, make_g2p
from ipatok import tokenise

AVAILABLE_G2P_ENGINES: dict[str, str | Callable] = {
    k: "DEFAULT_G2P" for k in get_arpabet_langs()[0]
}

# TODO: Add documentation for this
# If you want to override the default g2p engines, do so by the following:
#
# from some_cool_library import some_cool_g2p_method
# AVAILABLE_G2P_ENGINES['YOUR_LANGUAGE_CODE'] = some_cool_g2p_method
#
# IMPORTANT: Your g2p engine must return a list of tokenized symbols, and all of the returned symbols must be defined in your everyvoice-shared-text-config.yaml file.


class CachingG2PEngine:
    """caching tokenizing g2p engine"""

    def __init__(self, lang_id):
        self._cache = {}
        self.phonemizer = make_g2p(lang_id, f"{lang_id}-ipa")

    def process_one_token(self, input_token: str) -> list[str]:
        """Process one input token, dumbly split on whitespace.
        The output can be multiple tokens, since a proper tokenizer is used."""
        # ipatok strips some important characters, so as a hack,
        # we convert them to the private use area first
        PUA_CHARS = ["_", " ", ".", "ˈ", "ˌ"]
        PUA_START_NUMBER = 983040  # U+F0000
        text = self.phonemizer(input_token).output_string
        for i, char in enumerate(PUA_CHARS):
            text = text.replace(char, chr(PUA_START_NUMBER + i))
        tokens = tokenise(text, replace=False, tones=True, strict=False, unknown=True)
        # normalize the output since ipatok applies NFD
        unicode_normalization_form = self.phonemizer.transducers[-1].norm_form.value
        if unicode_normalization_form != "none":
            tokens = [normalize(unicode_normalization_form, token) for token in tokens]
        # convert the pua tokens back to their originals
        for i, token in enumerate(tokens):
            # PUA tokens have length 1
            if len(token) == 1:
                token_ord = ord(token)
                if token_ord >= PUA_START_NUMBER:
                    tokens[i] = PUA_CHARS[token_ord - PUA_START_NUMBER]
        return tokens

    def __call__(self, normalized_input_text: str) -> list[str]:
        input_tokens = re.split(r"(\s+)", normalized_input_text)
        output_tokens = []
        for token in input_tokens:
            cached = self._cache.get(token, None)
            if cached is None:
                cached = self.process_one_token(token)
                self._cache[token] = cached
            output_tokens += cached
        return output_tokens


def get_g2p_engine(lang_id: str):

    if lang_id not in AVAILABLE_G2P_ENGINES:
        raise NotImplementedError(
            f"Sorry, we don't have a grapheme-to-phoneme engine available for {lang_id}."
            " Please follow the docs to implement one yourself, or try training a character-based model instead."
        )

    if AVAILABLE_G2P_ENGINES[lang_id] == "DEFAULT_G2P":
        # Register the engine so we don't have to build it next time
        AVAILABLE_G2P_ENGINES[lang_id] = CachingG2PEngine(lang_id)

    return AVAILABLE_G2P_ENGINES[lang_id]



from enum import Enum
from pathlib import Path
from typing import Any, Optional

from everyvoice.config.preprocessing_config import PreprocessingConfig
from everyvoice.config.shared_types import (
    AdamOptimizer,
    AdamWOptimizer,
    BaseModelWithContact,
    BaseTrainingConfig,
    ConfigModel,
    init_context,
)
from everyvoice.config.text_config import TextConfig
from everyvoice.config.type_definitions import TargetTrainingTextRepresentationLevel
from everyvoice.config.utils import load_partials
from everyvoice.utils import load_config_from_json_or_yaml_path
from pydantic import Field, FilePath, ValidationInfo, field_serializer, model_validator


class DFAlignerExtractionMethod(Enum):
    beam = "beam"
    dijkstra = "dijkstra"


class DFAlignerModelConfig(ConfigModel):
    target_text_representation_level: TargetTrainingTextRepresentationLevel = (
        TargetTrainingTextRepresentationLevel.characters
    )
    lstm_dim: int = Field(
        512, description="The number of dimensions in the LSTM layers."
    )
    conv_dim: int = Field(
        512, description="The number of dimensions in the convolutional layers."
    )

    @field_serializer("target_text_representation_level")
    def convert_training_enum(
        self, target_text_representation_level: TargetTrainingTextRepresentationLevel
    ):
        return target_text_representation_level.value


class DFAlignerTrainingConfig(BaseTrainingConfig):
    optimizer: AdamOptimizer | AdamWOptimizer = Field(
        default_factory=AdamWOptimizer,
        description="Optimizer configuration settings.",
    )
    binned_sampler: bool = Field(True, description="Use a binned length sampler")
    plot_steps: int = Field(1000, description="The maximum number of steps to plot")
    extraction_method: DFAlignerExtractionMethod = Field(
        DFAlignerExtractionMethod.dijkstra,
        description="The alignment extraction algorithm to use. 'beam' will be quicker but possibly less accurate than 'dijkstra'",
    )

    @field_serializer("extraction_method")
    def convert_extraction_method_enum(
        self, extraction_method: DFAlignerExtractionMethod
    ):
        return extraction_method.value


class DFAlignerConfig(BaseModelWithContact):
    # TODO FastSpeech2Config and DFAlignerConfig are almost identical.
    model: DFAlignerModelConfig = Field(
        default_factory=DFAlignerModelConfig,
        description="The model configuration settings.",
    )
    path_to_model_config_file: Optional[FilePath] = Field(
        None, description="The path of a preprocessing configuration file."
    )

    training: DFAlignerTrainingConfig = Field(
        default_factory=DFAlignerTrainingConfig,
        description="The training configuration hyperparameters.",
    )
    path_to_training_config_file: Optional[FilePath] = Field(
        None, description="The path of a preprocessing configuration file."
    )

    preprocessing: PreprocessingConfig = Field(
        default_factory=PreprocessingConfig,
        description="The preprocessing configuration, including information about audio settings.",
    )
    path_to_preprocessing_config_file: Optional[FilePath] = Field(
        None, description="The path of a preprocessing configuration file."
    )

    text: TextConfig = Field(default_factory=TextConfig)
    path_to_text_config_file: Optional[FilePath] = None

    @model_validator(mode="before")  # type: ignore
    def load_partials(self: dict[Any, Any], info: ValidationInfo):
        config_path = (
            info.context.get("config_path", None) if info.context is not None else None
        )
        return load_partials(
            self,
            ("model", "training", "preprocessing", "text"),
            config_path=config_path,
        )

    @staticmethod
    def load_config_from_path(path: Path) -> "DFAlignerConfig":
        """Load a config from a path"""
        config = load_config_from_json_or_yaml_path(path)
        with init_context({"config_path": path}):
            config = DFAlignerConfig(**config)
        return config

    # INPUT_TODO: initialize text with union of symbols from dataset



from typing import Any, Dict

import pytorch_lightning as pl
from pytorch_lightning.callbacks import Callback
from typing_extensions import override


class ResetValidationDataloaderCallback(Callback):
    """
    Reset the validation progress to allow resuming and validating a full
    validation set and not just the first example in the validation set.
    """

    @override
    def on_save_checkpoint(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        checkpoint: Dict[str, Any],
    ) -> None:
        batch_progress = trainer.fit_loop.epoch_loop.val_loop.batch_progress
        batch_progress.reset()



import csv
import json
from collections import UserDict
from enum import Enum
from itertools import islice
from pathlib import Path
from typing import Dict, Iterable

import yaml
from tqdm import tqdm

from everyvoice.config.type_definitions import (
    DatasetTextRepresentation,
    TargetTrainingTextRepresentationLevel,
)


def rename_unknown_headers(headers):
    for i, header in enumerate(headers):
        if header not in ["basename", "raw_text", "speaker", "language"] + [
            x.value for x in DatasetTextRepresentation
        ]:
            headers[i] = f"unknown_{i}"
    return headers


def apply_automatic_text_conversions(
    filelist_data, text_representation, global_isocode=None
) -> str:
    """Arpabet is automatically converted to phones. phones are automatically derived from characters
       if a corresponding g2p module is available.

    Args:
        filelist_data (list[dict]): _description_
        text_representation (DatasetTextRepresentation): _description_
        isocode (str, optional): _description_. Defaults to None.
    Returns:
        target_training_representation (str): the target training representation level. returns 'phones' unless there is more data available from using 'characters'.
    """
    from everyvoice.text.arpabet import ARPABET_TO_IPA_TRANSDUCER
    from everyvoice.text.phonemizer import AVAILABLE_G2P_ENGINES, get_g2p_engine

    g2p_engines = {}
    character_counter = 0
    phone_counter = 0
    for item in tqdm(filelist_data, desc=f"Processing your {text_representation}"):
        # add globally defined language code
        if global_isocode is not None:
            item["language"] = global_isocode
            item_isocode = global_isocode
        else:
            # define the isocode from the item
            item_isocode = item.get("language", None)
        # convert arpabet to phones
        if (
            DatasetTextRepresentation.arpabet.value in item
            and DatasetTextRepresentation.ipa_phones.value not in item
        ):
            item[DatasetTextRepresentation.ipa_phones.value] = (
                ARPABET_TO_IPA_TRANSDUCER(
                    item[DatasetTextRepresentation.arpabet.value]
                ).output_string
            )
        # if phones don't exist but g2p is available, calculate them
        if (
            DatasetTextRepresentation.characters.value in item
            and DatasetTextRepresentation.ipa_phones.value not in item
            and item_isocode is not None
        ):
            if item_isocode in g2p_engines:
                pass  # just use the pre-loaded g2p engine.
            elif item_isocode in AVAILABLE_G2P_ENGINES:
                g2p_engines[item_isocode] = get_g2p_engine(item_isocode)
            else:
                g2p_engines[item_isocode] = None
            g2p_engine = g2p_engines[item_isocode]
            if g2p_engine is not None:
                phone_string_tokens = g2p_engine(
                    item[DatasetTextRepresentation.characters.value]
                )
                item[DatasetTextRepresentation.ipa_phones.value] = "".join(
                    phone_string_tokens
                )
        if DatasetTextRepresentation.ipa_phones.value in item:
            phone_counter += 1
        if DatasetTextRepresentation.characters.value in item:
            character_counter += 1
    if character_counter > phone_counter:
        target_training_representation = (
            TargetTrainingTextRepresentationLevel.characters.value
        )
    else:
        target_training_representation = (
            TargetTrainingTextRepresentationLevel.ipa_phones.value
        )
    if g2p_engines:
        for engine_iso, engine_value in g2p_engines.items():
            if engine_value is None:
                print(
                    f"Your text data contains characters and the language id {engine_iso} is not supported by a grapheme-to-phoneme engine. If you want to train using a pronunciation form for {engine_iso} (which usually results in better quality) you will have to add a g2p engine for {engine_iso}. Please see <TODO:Docs> for more information."
                )
    return target_training_representation


def read_unknown_tabular_filelist(
    path,
    delimiter=",",
    quoting=csv.QUOTE_NONE,
    escapechar="\\",
    record_limit: int = 0,  # if non-zero, read only this many records
) -> list[list[str]]:
    """Returns a list of list of cell values instead of a list of dicts as with
        the standard everyvoice.utils.generic_*sv_filelist_reader

    Returns:
        list[list[str]]: a list of rows containing a list of cell values
    """
    f: Iterable[str]
    with open(path, "r", newline="", encoding="utf8") as f:
        if record_limit:
            f = islice(f, record_limit)
        reader = csv.reader(
            f,
            delimiter=delimiter,
            quoting=quoting,
            escapechar=escapechar,
        )
        files = list(reader)
    return files


def write_dict_to_config(config: Dict, path: Path):
    """Given an object, write it to file.
       We have to serialize the json first to make use of the custom serializers,
       and we don't always write the json directly since we might want to add extra
       information after serialization.

    Args:
        config (Dict): The Configuration Dict to write
        path (Path): The output path; must end with either json or yaml
    """
    with open(path, "w", encoding="utf8") as f:
        path_string = str(path)
        if path_string.endswith("yaml"):
            yaml.dump(config, f, default_flow_style=None, allow_unicode=True)
        else:
            json.dump(config, f, ensure_ascii=False)


class EnumDict(UserDict):
    """dict that accepts Enum elements to mean their string values"""

    def convert_key(self, key):
        if isinstance(key, Enum):
            return key.value
        else:
            return key

    def __getitem__(self, key):
        return super().__getitem__(self.convert_key(key))

    def __setitem__(self, key, value):
        return super().__setitem__(self.convert_key(key), value)



import typer
from everyvoice.wizard import TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX

from .audit import audit as app_audit
from .benchmark import benchmark as app_benchmark
from .check_data import check_data as app_check_data
from .preprocess import preprocess as app_preprocess
from .synthesize import synthesize as app_synthesize
from .train import train as app_train

app = typer.Typer(
    pretty_exceptions_show_locals=False,
    context_settings={"help_option_names": ["-h", "--help"]},
    help="A PyTorch Lightning implementation of the FastSpeech2 Text-to-Speech Feature Prediction Model",
)

app.command(
    name="audit",
    short_help="",
)(app_audit)

app.command(
    name="benchmark",
    short_help="",
)(app_benchmark)

app.command(
    name="check_data",
    short_help="",
)(app_check_data)

app.command(
    name="preprocess",
    short_help="Preprocess your data",
    help=f"""
    # Preprocess Help

    This command will preprocess all of the data you need for use with EveryVoice.

    By default every step of the preprocessor will be done by running:
    \n\n
    **everyvoice preprocess config/{TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX}.yaml**
    \n\n
    If you only want to process specific things, you can run specific commands by adding them as options for example:
    \n\n
    **everyvoice preprocess config/{TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX}.yaml -s energy -s pitch**
    """,
)(app_preprocess)

app.command(
    name="synthesize",
    short_help="""Given some text and a trained model, generate some audio. i.e. perform typical speech synthesis""",
)(app_synthesize)

app.command(
    name="train",
    short_help="Train your Text-to-Spec model",
    help=f"""Train your text-to-spec model.  For example:

    **everyvoice train text-to-spec config/{TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX}.yaml**
    """,
)(app_train)



import json
from pathlib import Path

import typer
from everyvoice.base_cli.interfaces import (
    complete_path,
    load_config_base_command_interface,
)
from merge_args import merge_args


@merge_args(load_config_base_command_interface)
def check_data(
    filelist: Path = typer.Option(
        None,
        "--filelist",
        "-f",
        exists=True,
        dir_okay=False,
        file_okay=True,
        autocompletion=complete_path,
    ),
    **kwargs,
):
    from everyvoice.base_cli.helpers import load_config_base_command
    from everyvoice.preprocessor import Preprocessor
    from everyvoice.utils import generic_psv_filelist_reader

    from ..config import FastSpeech2Config

    config = load_config_base_command(
        model_config=FastSpeech2Config,
        **kwargs,
    )
    filelist = generic_psv_filelist_reader(filelist)
    preprocessor = Preprocessor(config)
    checked_data = preprocessor.check_data(filelist=filelist)
    with open("datapoints_sil_removed.json", "w", encoding="utf8") as f:
        json.dump(checked_data, f)



# Adapted from https://github.com/ufoym/imbalanced-dataset-sampler

from typing import Callable, Optional

import pandas as pd
import torch
import torch.utils.data


class ImbalancedDatasetSampler(torch.utils.data.sampler.Sampler):
    """Samples elements randomly from a given list of indices for imbalanced dataset

    Arguments:
        indices: a list of indices
        num_samples: number of samples to draw
        callback_get_label: a callback-like function which takes two arguments - dataset and index
    """

    def __init__(
        self,
        dataset,
        labels: Optional[list] = None,
        indices: Optional[list] = None,
        num_samples: Optional[int] = None,
        callback_get_label: Optional[Callable] = None,
    ):
        # if indices is not provided, all elements in the dataset will be considered
        self.indices = list(range(len(dataset))) if indices is None else indices

        # define custom callback
        self.callback_get_label = callback_get_label

        # if num_samples is not provided, draw `len(indices)` samples in each iteration
        self.num_samples = len(self.indices) if num_samples is None else num_samples

        # distribution of classes in the dataset
        df = pd.DataFrame()
        df["label"] = self._get_labels(dataset) if labels is None else labels
        df.index = self.indices
        df = df.sort_index()

        label_to_count = df["label"].value_counts()

        weights = 1.0 / label_to_count[df["label"]]
        self.weights = torch.DoubleTensor(weights.to_list())

    def _get_labels(self, dataset):
        if self.callback_get_label:
            return self.callback_get_label(dataset)
        elif isinstance(dataset, torch.utils.data.TensorDataset):
            return dataset.tensors[1]
        elif isinstance(dataset, torch.utils.data.Dataset):
            return dataset.get_labels()
        else:
            raise NotImplementedError

    def __iter__(self):
        return (
            self.indices[i]
            for i in torch.multinomial(self.weights, self.num_samples, replacement=True)
        )

    def __len__(self):
        return self.num_samples



from torch import nn


class DepthwiseSeparableConv1d(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int):
        super().__init__()
        self.model = nn.Sequential(
            nn.Conv1d(
                in_channels,
                in_channels,
                kernel_size,
                padding=(kernel_size - 1) // 2,
                groups=in_channels,
            ),
            nn.Conv1d(in_channels, out_channels, 1),
        )

    def forward(self, x):
        return self.model(x)


class LayerNorm(nn.LayerNorm):
    """Layer normalization module.
    :param int nout: output dim size
    :param int dim: dimension to be normalized
    """

    def __init__(self, nout, dim=-1):
        """Construct an LayerNorm object."""
        super(LayerNorm, self).__init__(nout, eps=1e-12)
        self.dim = dim

    def forward(self, x):
        """Apply layer normalization.
        :param torch.Tensor x: input tensor
        :return: layer normalized tensor
        :rtype torch.Tensor
        """
        if self.dim == -1:
            return super(LayerNorm, self).forward(x)
        return super(LayerNorm, self).forward(x.transpose(1, -1)).transpose(1, -1)


class ConvNorm(nn.Module):
    """1D Convolution"""

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=1,
        stride=1,
        padding=None,
        dilation=1,
        bias=True,
        w_init_gain="linear",
        transpose=False,
    ):
        super(ConvNorm, self).__init__()

        if padding is None:
            assert kernel_size % 2 == 1
            padding = int(dilation * (kernel_size - 1) / 2)

        self.conv = nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=bias,
        )

        nn.init.xavier_uniform_(
            self.conv.weight, gain=nn.init.calculate_gain(w_init_gain)
        )
        self.transpose = transpose

    def forward(self, x):
        if self.transpose:
            x = x.contiguous().transpose(1, 2)
        x = self.conv(x)
        if self.transpose:
            x = x.contiguous().transpose(1, 2)

        return x



import json
from enum import Enum

import typer
from everyvoice.base_cli.interfaces import preprocess_base_command_interface
from merge_args import merge_args


class PreprocessCategories(str, Enum):
    audio = "audio"
    spec = "spec"
    attn = "attn"
    text = "text"
    pitch = "pitch"
    energy = "energy"


@merge_args(preprocess_base_command_interface)
def preprocess(
    compute_stats: bool = typer.Option(
        True, "-S", "--stats", help="Calculate stats for energy and pitch"
    ),
    steps: list[PreprocessCategories] = typer.Option(
        [cat.value for cat in PreprocessCategories],
        "-s",
        "--steps",
        help="Which steps of the preprocessor to use. If none are provided, all steps will be performed.",
    ),
    **kwargs,
):
    from everyvoice.base_cli.helpers import preprocess_base_command

    from ..config import FastSpeech2Config

    preprocessor, config, processed = preprocess_base_command(
        model_config=FastSpeech2Config,
        steps=[step.name for step in steps],
        **kwargs,
    )

    if compute_stats:
        stats_path = config.preprocessing.save_dir / "stats.json"
        e_scaler, p_scaler = preprocessor.compute_stats(
            energy="energy" in processed, pitch="pitch" in processed
        )
        stats = {}
        if e_scaler:
            e_stats = e_scaler.calculate_stats()
            stats["energy"] = e_stats
        if p_scaler:
            p_stats = p_scaler.calculate_stats()
            stats["pitch"] = p_stats
        preprocessor.normalize_stats(e_scaler, p_scaler)
        # Merge with existing stats
        if stats_path.exists():
            with open(stats_path, "r", encoding="utf8") as f:
                previous_stats = json.load(f)
        else:
            previous_stats = {}
        stats = {**previous_stats, **stats}
        with open(
            config.preprocessing.save_dir / "stats.json", "w", encoding="utf8"
        ) as f:
            json.dump(stats, f)



from torch.nn import Conv1d, ConvTranspose1d, Sequential
from torch.nn.utils import weight_norm as wnorm


def create_depthwise_separable_convolution(
    in_channels,
    out_channels,
    kernel_size,
    stride=1,
    padding=0,
    output_padding=0,
    dilation=1,
    groups=1,
    bias=True,
    padding_mode="zeros",
    transpose=False,
    weight_norm=True,
):
    if transpose:
        depth = ConvTranspose1d(
            in_channels,
            in_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
            dilation=dilation,
            bias=bias,
            groups=in_channels,
        )
        point = ConvTranspose1d(in_channels, out_channels, kernel_size=1)
    else:
        depth = Conv1d(
            in_channels,
            in_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
            dilation=dilation,
            bias=bias,
            groups=in_channels,
        )
        point = Conv1d(in_channels, out_channels, kernel_size=1)
    if weight_norm:
        return Sequential(wnorm(depth), wnorm(point))
    else:
        return Sequential(depth, point)



import tempfile
from pathlib import Path
from string import ascii_lowercase

from everyvoice.config.preprocessing_config import Dataset, PreprocessingConfig
from everyvoice.config.text_config import Symbols, TextConfig
from everyvoice.model.e2e.config import FeaturePredictionConfig
from everyvoice.preprocessor import Preprocessor
from everyvoice.tests.basic_test_case import BasicTestCase
from everyvoice.utils import collapse_whitespace, lower, nfc_normalize


class PreprocessedAudioFixture:
    """
    A unittest fixture to preprocess the audio files.
    """

    _tempdir = tempfile.TemporaryDirectory(prefix="tmpdir_PreprocessedInputFixture_")
    lj_preprocessed = Path(_tempdir.name)
    _preprocess_ran = False

    data_dir = BasicTestCase.data_dir
    wavs_dir = data_dir / "lj" / "wavs"
    lj_filelist = lj_preprocessed / "preprocessed_filelist.psv"

    fp_config = FeaturePredictionConfig(
        preprocessing=PreprocessingConfig(
            save_dir=lj_preprocessed,
            source_data=[
                Dataset(
                    data_dir=wavs_dir,
                    filelist=data_dir / "metadata.psv",
                    permissions_obtained=True,
                )
            ],
        ),
        text=TextConfig(
            cleaners=[collapse_whitespace, lower, nfc_normalize],
            symbols=Symbols(
                ascii_symbols=list(ascii_lowercase),
                ipa=[
                    "ɔ",
                    "æ",
                    "ɡ",
                    "ɛ",
                    "ð",
                    "ɜ˞",
                    "ʌ",
                    "ɑ",
                    "ɹ",
                    "ʃ",
                    "ɪ",
                    "ʊ",
                    "ʒ",
                ],
            ),
        ),
        contact=BasicTestCase.contact,
    )

    preprocessor = Preprocessor(fp_config)

    @classmethod
    def setUpClass(cls):
        """Generate a preprocessed test set that can be used in various test cases."""
        # We only need to actually run this once
        if not PreprocessedAudioFixture._preprocess_ran:
            PreprocessedAudioFixture.preprocessor.preprocess(
                output_path=str(PreprocessedAudioFixture.lj_filelist),
                cpus=1,
                overwrite=True,
                to_process=("audio", "energy", "pitch", "text", "spec"),
            )
            PreprocessedAudioFixture.lj_preprocessed.mkdir(parents=True, exist_ok=True)
            (PreprocessedAudioFixture.lj_preprocessed / "duration").symlink_to(
                PreprocessedAudioFixture.data_dir / "lj" / "preprocessed" / "duration",
            )

            PreprocessedAudioFixture._preprocess_ran = True



import numpy as np
from scipy.sparse import coo_matrix
from scipy.sparse.csgraph import dijkstra


def to_node_index(i, j, cols):
    return cols * i + j


def from_node_index(node_index, cols):
    return node_index // cols, node_index % cols


def to_adj_matrix(mat):
    rows = mat.shape[0]
    cols = mat.shape[1]

    row_ind = []
    col_ind = []
    data = []

    for i in range(rows):
        for j in range(cols):

            node = to_node_index(i, j, cols)

            if j < cols - 1:
                right_node = to_node_index(i, j + 1, cols)
                weight_right = mat[i, j + 1]
                row_ind.append(node)
                col_ind.append(right_node)
                data.append(weight_right)

            if i < rows - 1 and j < cols:
                bottom_node = to_node_index(i + 1, j, cols)
                weight_bottom = mat[i + 1, j]
                row_ind.append(node)
                col_ind.append(bottom_node)
                data.append(weight_bottom)

            if i < rows - 1 and j < cols - 1:
                bottom_right_node = to_node_index(i + 1, j + 1, cols)
                weight_bottom_right = mat[i + 1, j + 1]
                row_ind.append(node)
                col_ind.append(bottom_right_node)
                data.append(weight_bottom_right)

    adj_mat = coo_matrix((data, (row_ind, col_ind)), shape=(rows * cols, rows * cols))
    return adj_mat.tocsr()


def extract_durations_with_dijkstra(tokens: np.ndarray, pred: np.ndarray) -> np.ndarray:
    """
    Extracts durations from the attention matrix by finding the shortest monotonic path from
    top left to bottom right.
    """

    pred_max = pred[:, tokens]
    path_probs = 1.0 - pred_max
    adj_matrix = to_adj_matrix(path_probs)
    dist_matrix, predecessors = dijkstra(
        csgraph=adj_matrix, directed=True, indices=0, return_predecessors=True
    )
    path = []
    pr_index = predecessors[-1]
    while pr_index != 0:
        path.append(pr_index)
        pr_index = predecessors[pr_index]
    path.reverse()

    # append first and last node
    path = [0] + path + [dist_matrix.size - 1]
    cols = path_probs.shape[1]
    mel_text = {}
    durations = np.zeros(tokens.shape[0], dtype=np.int32)

    # collect indices (mel, text) along the path
    for node_index in path:
        i, j = from_node_index(node_index, cols)
        mel_text[i] = j

    for j in mel_text.values():
        durations[j] += 1

    return durations


def extract_durations_beam(
    tokens: np.ndarray, pred: np.ndarray, k: int
) -> tuple[list[np.ndarray], list[list[np.ndarray]]]:
    data = pred[:, tokens]
    sequences = [[[0], -np.log(data[0, 0])]]  # always start on first position
    for row in data[1:]:
        all_candidates = []
        # expand each current candidate
        for i in range(len(sequences)):
            seq, score = sequences[i]
            for j in [seq[-1], seq[-1] + 1]:  # only allow 2 possible moves
                if j < data.shape[-1]:
                    candidate = [seq + [j], score - np.log(row[j])]
                else:
                    candidate = [seq + [j], np.inf]
                all_candidates.append(candidate)
        # order all candidates by score
        ordered = sorted(all_candidates, key=lambda tup: tup[1])
        # select k best
        sequences = ordered[:k]
    durations = [np.bincount(sequence[0]) for sequence in sequences]
    return durations, sequences



import doctest
import tempfile
from pathlib import Path
from typing import Any
from unittest import TestCase

import torch
from pep440 import is_canonical
from pydantic import BaseModel
from pydantic.functional_validators import BeforeValidator
from typing_extensions import Annotated

import everyvoice.utils
from everyvoice._version import VERSION
from everyvoice.config.shared_types import _init_context_var, init_context
from everyvoice.config.validation_helpers import (
    directory_path_must_exist,
    path_is_a_directory,
    relative_to_absolute_path,
)
from everyvoice.tests.stubs import capture_logs
from everyvoice.utils import write_filelist
from everyvoice.utils.heavy import get_device_from_accelerator


class VersionTest(TestCase):
    def test_version_is_pep440_compliant(self):
        self.assertTrue(is_canonical(VERSION))


class UtilsTest(TestCase):
    def test_run_doctest(self):
        """Run doctests in everyvoice.utils"""
        results = doctest.testmod(everyvoice.utils)
        self.assertFalse(results.failed, results)

    def test_write_filelist(self):
        """Filelist should write files with headers in order"""
        basic_files = [
            {
                "basename": "test",
                "phones": "foo",
                "characters": "bar",
                "language": "test",
                "extra": "test",
            }
        ]
        with tempfile.TemporaryDirectory() as tempdir:
            tempdir = Path(tempdir)
            basic_path = tempdir / "test.psv"
            write_filelist(basic_files, basic_path)
            with open(basic_path) as f:
                headers = f.readline().strip().split("|")
            self.assertEqual(len(headers), 5)
            self.assertEqual(headers[0], "basename")
            self.assertEqual(headers[1], "language")
            self.assertEqual(headers[2], "characters")
            self.assertEqual(headers[3], "phones")
            self.assertEqual(headers[4], "extra")


class ContextableBaseModel(BaseModel):
    """
    Enable using
       with init_context({"k":v}):
          pass
    which enable passing a context down to a pydantic BaseModel.
    """

    # [Using validation context with BaseModel initialization](https://docs.pydantic.dev/2.3/usage/validators/#using-validation-context-with-basemodel-initialization)
    def __init__(__pydantic_self__, **data: Any) -> None:
        __pydantic_self__.__pydantic_validator__.validate_python(
            data,
            self_instance=__pydantic_self__,
            context=_init_context_var.get(),
        )


class PathIsADirectory(ContextableBaseModel):
    """Dummy Class for PathIsADirectoryTest"""

    path: Annotated[Path, BeforeValidator(path_is_a_directory)]


class PathIsADirectoryTest(TestCase):
    """Testing when we Annotated with path_is_a_directory"""

    def test_using_a_directory_with_context(self):
        """
        Verifies that PathIsADirectory detects that the argument is a directory
        when using a context.
        """
        try:
            root_dir = Path(__file__).parent / "data"
            root_dir = root_dir.resolve()
            directory = Path("hierarchy")
            self.assertTrue((root_dir / directory).exists())
            with init_context({"writing_config": root_dir}):
                PathIsADirectory(path=directory)
        except ValueError:
            self.fail("Failed to detect that the argument is a directory")

    def test_using_a_file_with_context(self):
        """
        Verifies that PathIsADirectory detects that the argument is a file when
        using a context.
        """
        file = Path(__file__)
        with self.assertRaisesRegex(
            ValueError,
            rf"{file} is not a directory",
        ):
            with init_context({"writing_config": file.parent.resolve()}):
                PathIsADirectory(path=file.name)

    def test_invalid_type(self):
        """The argument to PathIsADirectory must be of type Path"""
        with self.assertRaisesRegex(
            ValueError,
            r".*Value error,  \[type=value_error, input_value=4, input_type=int\].*",
        ):
            PathIsADirectory(path=4)

    def test_using_a_directory(self):
        """
        Verifies that PathIsADirectory detects that the argument is a directory.
        """
        try:
            root_dir = Path(__file__).parent / "data"
            root_dir = root_dir.resolve()
            directory = Path("hierarchy")
            self.assertTrue((root_dir / directory).exists())
            PathIsADirectory(path=root_dir / directory)
        except ValueError:
            self.fail("Failed to detect that the argument is a directory")

    def test_using_a_file(self):
        """
        Verifies that PathIsADirectory detects that the argument is a file.
        """
        path = Path(__file__)
        with self.assertRaisesRegex(
            ValueError,
            rf"{path} is not a directory",
        ):
            PathIsADirectory(path=path)


class RelativePathToAbsolute(ContextableBaseModel):
    """Dummy Class for RelativePathToAbsoluteTest"""

    path: Annotated[Path | None, BeforeValidator(relative_to_absolute_path)]


class RelativePathToAbsoluteTest(TestCase):
    """Testing when we Annotated with relative_to_absolute_path"""

    def test_None(self):
        """A special case, if path is None, it should stay None"""
        dir = RelativePathToAbsolute(path=None)
        self.assertIsNone(dir.path)

    def test_invalid_type(self):
        """
        If the provided type cannot be a path, it should fail.
        """
        with self.assertRaisesRegex(
            ValueError,
            r".*Value error,  \[type=value_error, input_value=4, input_type=int\].*",
        ):
            RelativePathToAbsolute(path=4)

    def test_already_absolute(self):
        """Under a init_context, an absolute path must stay the same"""
        path = Path(__file__).absolute()
        with init_context({"config_path": path.parent / "data"}):
            dir = RelativePathToAbsolute(path=path)
            self.assertEqual(dir.path, path)

    def test_should_not_change(self):
        """
        Without context, the path should stay the same.
        """
        path = Path("data")
        test = RelativePathToAbsolute(path=path)
        self.assertEqual(test.path, path)

    def test_with_context(self):
        """
        When provided with a context, the path should be absolute.
        """
        root_dir = Path(__file__).parent.resolve()
        path = Path("data")
        with init_context({"config_path": root_dir}):
            dir = RelativePathToAbsolute(path=path)
            self.assertTrue(dir.path.is_absolute())


class DirectoryPathMustExist(ContextableBaseModel):
    """Dummy Class for DirectoryPathMustExistTest"""

    path: Annotated[Path, BeforeValidator(directory_path_must_exist)]


class DirectoryPathMustExistTest(TestCase):
    """
    Testing when we Annotated with directory_path_must_exist.
    It should create a directory if it doesn't exist.
    """

    def test_argument_is_Path(self):
        """directory_path_must_exist() expects a Path"""
        with self.assertRaisesRegex(
            ValueError,
            r".*Assertion failed,  \[type=assertion_error, input_value='invalid_not_a_Path', input_type=str\].*",
        ):
            DirectoryPathMustExist(path="invalid_not_a_Path")

    def test_using_a_directory_with_context(self):
        """
        Verifies that directory_path_must_exist(), when using a context,
        creates the directory if it doesn't exist.
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            root_dir = Path(tmpdir)
            directory = Path("test_using_a_directory_with_context")
            self.assertFalse((root_dir / directory).exists())
            with init_context({"writing_config": root_dir.resolve()}):
                dir = DirectoryPathMustExist(path=directory)
            # Note: dir.path shouldn't not change to an absolute value.
            self.assertEqual(dir.path, directory)
            self.assertTrue((root_dir / directory).exists())
            # Note: since dir.path is NOT replaced with an absolute it
            # shouldn't exist because it was created relative to the context's
            # path.
            self.assertFalse(dir.path.exists())

    def test_path_already_exists(self):
        path = Path(__file__).parent / "data"
        with capture_logs() as output:
            dir = DirectoryPathMustExist(path=path)
            # There should be no info logged.
            self.assertListEqual(output, [])
        self.assertTrue(dir.path.exists())

    def test_using_a_directory(self):
        """
        Automatically create a directory.
        """
        with tempfile.TemporaryDirectory() as tmpdir, capture_logs() as output:
            path = Path(tmpdir) / "test_using_a_directory"
            self.assertFalse(path.exists())
            dir = DirectoryPathMustExist(path=path)
            self.assertEqual(dir.path, path)
            self.assertIn(
                f"Directory at {path} does not exist. Creating...",
                output[0],
            )
            self.assertTrue(path.exists())
            self.assertTrue(dir.path.exists())


class GetDeviceFromAcceleratorTest(TestCase):
    def test_auto(self):
        self.assertEqual(
            get_device_from_accelerator("auto"),
            torch.device("cuda:0" if torch.cuda.is_available() else "cpu"),
        )

    def test_cpu(self):
        self.assertEqual(get_device_from_accelerator("cpu"), torch.device("cpu"))

    def test_gpu(self):
        self.assertEqual(get_device_from_accelerator("gpu"), torch.device("cuda:0"))

    def test_mps(self):
        self.assertEqual(get_device_from_accelerator("mps"), torch.device("mps"))

    def test_unknown_accelerator(self):
        self.assertEqual(get_device_from_accelerator("unknown"), torch.device("cpu"))



import json
import numbers
import tempfile
from enum import Enum
from pathlib import Path

import torch
from pytorch_lightning import Trainer

from everyvoice.config.type_definitions import DatasetTextRepresentation
from everyvoice.model.aligner.DeepForcedAligner.dfaligner.config import DFAlignerConfig
from everyvoice.model.aligner.DeepForcedAligner.dfaligner.model import Aligner
from everyvoice.model.e2e.config import EveryVoiceConfig
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.config import (
    FastSpeech2Config,
)
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.model import (
    FastSpeech2,
)
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.type_definitions_heavy import (
    Stats,
    StatsInfo,
)
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.config import HiFiGANConfig
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.model import HiFiGAN
from everyvoice.tests.basic_test_case import BasicTestCase
from everyvoice.tests.stubs import monkeypatch
from everyvoice.wizard import (
    ALIGNER_CONFIG_FILENAME_PREFIX,
    SPEC_TO_WAV_CONFIG_FILENAME_PREFIX,
    TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX,
)

monkey_saved_config = None


def find_non_basic_substructures(structure):
    """Return a list of all substructures within structure that are not basic data types.

    Dict, tuple, list, string, numbers are considered basic, stuff that depends on our
    code definitions are not.

    We use this function to make sure our checkpoints are robust to code refactorings.
    If we refactor and move the code for a class to a different file for performance
    or clarity reasons, but that class is embeded in a checkpoint, that checkpoint
    will no longer be loadable, so we want to block that early.
    """
    if isinstance(structure, Enum):
        return [structure]
    if isinstance(structure, (type(None), str, numbers.Number)):
        return []
    if isinstance(structure, (list, tuple)):
        result = []
        for element in structure:
            result.extend(find_non_basic_substructures(element))
        return result
    if isinstance(structure, dict):
        result = []
        for key, value in structure.items():
            result.extend(find_non_basic_substructures(key))
            result.extend(find_non_basic_substructures(value))
        return result
    return [structure]


class ModelTest(BasicTestCase):
    """Basic test for models"""

    def setUp(self) -> None:
        super().setUp()
        self.config = EveryVoiceConfig(
            contact=self.contact,
        )
        self.config_dir = self.data_dir / "relative" / "config"

    def test_hparams(self):
        self.hifi_gan = HiFiGAN(self.config.vocoder)
        self.assertEqual(self.config.vocoder, self.hifi_gan.hparams.config)
        self.assertEqual(self.config.vocoder, self.hifi_gan.config)

    def test_checkpoints_only_contain_serializable_content(self):
        """These tests help remove any dependencies on specific versions of Pydantic.
        By serializing our checkpoint hyperparameters and configuration with only JSON objects,
        we can help allow our models to be loaded by other versions of EveryVoice. This test ensures
        the hyperparameters only contain JSON serializable content
        """
        SERIAL_SAFE_MODELS = [
            HiFiGAN(
                HiFiGANConfig.load_config_from_path(
                    self.config_dir / f"{SPEC_TO_WAV_CONFIG_FILENAME_PREFIX}.yaml"
                )
            ),
            FastSpeech2(
                FastSpeech2Config.load_config_from_path(
                    self.config_dir / f"{TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX}.yaml"
                ),
                stats=Stats(
                    pitch=StatsInfo(
                        min=0, max=1, std=2, mean=3, norm_min=4, norm_max=5
                    ),
                    energy=StatsInfo(
                        min=7, max=8, std=9, mean=10, norm_min=11, norm_max=12
                    ),
                ),
                lang2id={"foo": 0, "bar": 1},
                speaker2id={"baz": 0, "qux": 1},
            ),  # we should probably also test that the error about the variance adaptor is raised
            Aligner(
                DFAlignerConfig.load_config_from_path(
                    self.config_dir / f"{ALIGNER_CONFIG_FILENAME_PREFIX}.yaml"
                )
            ),
        ]
        for model in SERIAL_SAFE_MODELS:
            trainer = Trainer()
            with tempfile.TemporaryDirectory() as tmpdir_str:
                # Hacky way to connect the trainer with a model instead of trainer.fit(model) just for testing
                # https://lightning.ai/forums/t/saving-a-lightningmodule-without-a-trainer/2217/2
                trainer.strategy.connect(model)
                tmpdir = Path(tmpdir_str)

                def monkey_on_save_checkpoint(checkpoint):
                    ret = model.real_on_save_checkpoint(checkpoint)
                    global monkey_saved_config
                    monkey_saved_config = checkpoint["hyper_parameters"]["config"]
                    return ret

                with monkeypatch(
                    model, "real_on_save_checkpoint", model.on_save_checkpoint
                ):
                    with monkeypatch(
                        model, "on_save_checkpoint", monkey_on_save_checkpoint
                    ):
                        trainer.save_checkpoint(tmpdir / "model.ckpt")

                result = find_non_basic_substructures(monkey_saved_config)
                self.assertEqual(
                    result,
                    [],
                    f"The following parts of the checkpoint were not properly serialized by on_model_checkpoint(): {result}",
                )

                # We don't want just serializable, but actually serialized!
                ckpt = torch.load(tmpdir / "model.ckpt")
                try:
                    json.dumps(ckpt["hyper_parameters"])
                except (TypeError, OverflowError):
                    self.fail(
                        f"model {type(model).__name__} has some fields that are not JSON serializable"
                    )

    def test_find_non_basic_substructures(self):
        """Let's make sure the testing facility itself works..."""
        self.assertEqual(
            find_non_basic_substructures([{1: 2, "a": [3, 2]}, (1 / 3, 2.0), None]), []
        )
        si = StatsInfo(min=0, max=1, std=2, mean=3, norm_min=4, norm_max=5)
        result = find_non_basic_substructures(
            [
                {1: 2, "a": [3, 2]},
                (1 / 3, 2.0),
                Stats,
                {"a": si},
                {DatasetTextRepresentation.characters: "b"},
            ]
        )
        self.assertEqual(result, [Stats, si, DatasetTextRepresentation.characters])



"""Preprocessor Module that given a filelist containing text, wav, textgrid:
- extracts log Mel spectral features
- extracts pitch (phone-level or frame-level)
- extracts energy
- extracts inputs (ex. phonological feats)
"""

import functools
import json
import multiprocessing as mp
import random
import sys
from collections import Counter
from glob import glob
from multiprocessing import Manager
from pathlib import Path
from typing import Optional

import numpy as np
import numpy.typing as npt
import torch
import torchaudio
from clipdetect import detect_clipping
from joblib import Parallel, delayed
from loguru import logger
from rich import print as rich_print
from rich.panel import Panel
from rich.style import Style
from tabulate import tabulate
from torchaudio.functional import resample
from torchaudio.sox_effects import apply_effects_tensor
from tqdm import tqdm

from everyvoice.config.type_definitions import (
    DatasetTextRepresentation,
    TargetTrainingTextRepresentationLevel,
)
from everyvoice.exceptions import ConfigError
from everyvoice.model.aligner.config import AlignerConfig
from everyvoice.model.feature_prediction.config import FeaturePredictionConfig
from everyvoice.model.vocoder.config import VocoderConfig
from everyvoice.preprocessor.attention_prior import BetaBinomialInterpolator
from everyvoice.preprocessor.helpers import Counters, Scaler, save_tensor, save_wav
from everyvoice.text.arpabet import ARPABET_TO_IPA_TRANSDUCER
from everyvoice.text.phonemizer import AVAILABLE_G2P_ENGINES
from everyvoice.text.text_processor import TextProcessor
from everyvoice.utils import (
    generic_psv_filelist_reader,
    n_times,
    tqdm_joblib_context,
    write_filelist,
)
from everyvoice.utils.heavy import (
    dynamic_range_compression_torch,
    get_spectral_transform,
)


class Preprocessor:
    def __init__(self, config: AlignerConfig | FeaturePredictionConfig | VocoderConfig):
        self.config = config
        self.counters = Counters(Manager())
        self.cpus = 0
        self.pitch_scaler = Scaler()
        self.energy_scaler = Scaler()
        self.datasets = config.preprocessing.source_data
        self.save_dir = Path(config.preprocessing.save_dir)
        self.audio_config = config.preprocessing.audio
        self.sep = "--"
        self.text_processor = (
            None if isinstance(config, VocoderConfig) else TextProcessor(config.text)
        )
        self.overwrite = False
        self.input_sampling_rate = self.audio_config.input_sampling_rate
        self.output_sampling_rate = self.audio_config.output_sampling_rate
        self.sampling_rate_change = (
            self.output_sampling_rate // self.input_sampling_rate
        )
        self.output_hop_size = (
            self.audio_config.fft_hop_size * self.sampling_rate_change
        )
        # Define Spectral Transform
        # Gah, so many ways to do this: https://github.com/CookiePPP/VocoderComparisons/issues/3
        self.input_spectral_transform = get_spectral_transform(
            self.audio_config.spec_type,
            self.audio_config.n_fft,
            self.audio_config.fft_window_size,
            self.audio_config.fft_hop_size,
            sample_rate=self.input_sampling_rate,
            n_mels=self.audio_config.n_mels,
            f_min=self.audio_config.f_min,
            f_max=self.audio_config.f_max,
        )
        self.output_spectral_transform = get_spectral_transform(
            self.audio_config.spec_type,
            self.audio_config.n_fft * self.sampling_rate_change,
            self.audio_config.fft_window_size * self.sampling_rate_change,
            self.output_hop_size,
            sample_rate=self.input_sampling_rate,
            n_mels=self.audio_config.n_mels,
            f_min=self.audio_config.f_min,
            f_max=self.audio_config.f_max,
        )

        if (
            self.input_spectral_transform is None
            or self.output_spectral_transform is None
        ):
            raise ConfigError(
                f"Spectral feature specification '{self.audio_config.spec_type}' is not supported. Please edit your config file."
            )

    def load_audio(self, audio_path: Path) -> tuple[torch.Tensor, int, float]:
        """
        Load an audio file and calculate its duration.

        Args:
            audio_path: path to audio file

        Returns: (audio as a Tensor, sampling rate, duration in seconds)
        """
        audio, sr = torchaudio.load(str(audio_path))
        seconds = len(audio[0]) / sr
        return audio, sr, seconds

    def process_audio(
        self,
        wav_path: Path,
        normalize=True,
        resample_rate=None,
        sox_effects=None,
        hop_size=None,
        update_counters=True,  # unset this when processing the same file a second time
    ) -> tuple[torch.Tensor, int] | tuple[None, None]:
        """Process audio

        Args:
            wav_path (Path): path to wav file
            normalize (bool): volume normalization
        Returns:
            Tensor: (audio as a Tensor, sampling rate)
        """
        audio, sr, seconds = self.load_audio(wav_path)

        if seconds > self.audio_config.max_audio_length:
            logger.warning(
                f"Audio too long: {wav_path} ({seconds} seconds - we will skip this file)"
            )
            if update_counters:
                self.counters.increment("audio_too_long")
            return None, None
        if seconds < self.audio_config.min_audio_length:
            logger.warning(
                f"Audio too short: {wav_path} ({seconds} seconds - we will skip this file)"
            )
            if update_counters:
                self.counters.increment("audio_too_short")
            return None, None

        loudness_transform = torchaudio.transforms.Loudness(sr)
        loudness = loudness_transform(audio)
        if (
            torch.isnan(loudness) or loudness < -36
        ):  # This is a conservative threshold, so some very quiet/silent files may still get through
            logger.warning(f"Audio empty: {wav_path} - we will skip this file")
            if update_counters:
                self.counters.increment("audio_empty")
            return None, None

        if sox_effects:
            audio, sr = apply_effects_tensor(
                audio,
                sr,
                sox_effects,
            )

        if resample_rate is not None and resample_rate != sr:
            audio = resample(audio, sr, resample_rate)
            sr = resample_rate
        if normalize:
            audio /= torch.max(torch.abs(audio))
            audio *= 0.95

        if update_counters:
            self.counters.increment("processed_files")
            self.counters.increment("duration", seconds)

        audio = audio.squeeze()  # get rid of channels dimension
        # limit the number of samples to a number that is evenly divisible by the hop size
        # most reasonable hop sizes for training should be in the vicinity of ~10ms
        # so we're talking about losing slightly less than that amount of audio from the
        # signal at a maximum, and it makes downstream things like vocoder matching more straightforward.
        if hop_size is None:
            raise ValueError(
                "We must know the hop size for processing audio because EveryVoice enforces that the number of samples is evenly divisible by the hop size"
            )
        max_frames = audio.size(0) // hop_size
        max_samples = max_frames * hop_size
        return audio[:max_samples], sr

    def extract_spectral_features(
        self, audio_tensor: torch.Tensor, transform, normalize=True
    ):
        """Given an audio tensor, extract the log Mel spectral features
        from the given start and end points

        Args:
            audio_tensor (Tensor): Tensor trimmed according
            transform (torchaudio.transforms): transform to apply; use either Preprocessor.input_spectral_transform or Preprocessor.output_spectral_transform
        """
        mel = transform(audio_tensor)
        if normalize:
            mel = dynamic_range_compression_torch(mel)
        return mel

    @staticmethod
    def _interpolate(x):
        def nan_helper(y):
            return np.isnan(y), lambda z: z.nonzero()[0]

        nans, y = nan_helper(x)
        x[nans] = np.interp(y(nans), y(~nans), x[~nans])
        return x

    def extract_pitch(self, audio_tensor: torch.Tensor):
        """Given an audio tensor, extract the pitch

        TODO: consider CWT and Parselmouth

        Comparison with other implementations:
            - ming024 & Christoph Minxhoffer use the pyworld implementation and interpolate along with phone averaging
            - the Lightspeech implementation seems to use pyworld implementation and not interpolate or average
            - Christoph Minxhoffer reported no significant differences with continuous wavelet transform so it is not implemented here

        Args:
            audio_tensor (Tensor): 1D tensor of audio samples
        """
        import pyworld as pw  # This isn't a very good place for an import,

        # but also pyworld is very annoying to install so this is a compromise
        pitch, t = pw.dio(
            audio_tensor.squeeze(0)
            .numpy()
            .astype(
                np.float64
            ),  # TODO: why are these np.float64, maybe it's just what pw expects?
            self.input_sampling_rate,
            frame_period=self.audio_config.fft_hop_size
            / self.input_sampling_rate
            * 1000,
            speed=4,
        )
        pitch = pw.stonemask(
            audio_tensor.squeeze(0).numpy().astype(np.float64),
            pitch,
            t,
            self.input_sampling_rate,
        )
        pitch[pitch == 0] = np.nan
        try:
            pitch = self._interpolate(pitch)
        except ValueError:
            # TODO: we should warn the user about pitch-less samples
            pitch[np.isnan(pitch)] = 0
        pitch = torch.tensor(pitch).float()
        return pitch

    def average_data_by_durations(self, data, durations):
        current_frame_position = 0
        new_data = []
        for duration in durations.numpy().tolist():
            if duration > 0:
                new_data.append(
                    torch.mean(
                        data[current_frame_position : current_frame_position + duration]
                    )
                )
            else:
                new_data.append(1e-7)
            current_frame_position += duration
        return torch.tensor(new_data)

    def extract_energy(self, spectral_feature_tensor: torch.Tensor):
        """Given a spectral feature tensor, and durations extract the energy averaged across a phone

        Args:
            spectral_feature_tensor (Tensor): tensor of spectral features extracted from audio
            durations (_type_): _descriptiont    #TODO
        """
        return torch.linalg.norm(spectral_feature_tensor, dim=0)

    def print_duration(self):
        """Convert seconds to a human readable format"""
        seconds = int(self.counters.value("duration"))
        hours = seconds // 3600
        seconds %= 3600
        minutes = seconds // 60
        seconds %= 60
        return f"{hours}h {minutes}m {seconds}s"

    def report(self, tablefmt="simple"):
        """Print a report of the dataset processing"""
        headers = ["type", "quantity"]
        table = [
            ["processed files", self.counters.value("processed_files")],
            [
                "previously processed files",
                self.counters.value("previously_processed_files"),
            ],
            ["missing files", self.counters.value("missing_files")],
            [
                "missing symbols",
                len(self.text_processor.missing_symbols) if self.text_processor else 0,
            ],
            ["skipped processes", self.counters.value("skipped_processes")],
            ["nans", self.counters.value("nans")],
            ["audio_empty", self.counters.value("audio_empty")],
            ["audio_too_short", self.counters.value("audio_too_short")],
            ["audio_too_long", self.counters.value("audio_too_long")],
            ["duration", self.print_duration()],
        ]
        return tabulate(table, headers, tablefmt=tablefmt)

    def get_speaker_and_language(self, item):
        """Unless the dataset already has values for speaker and language, set them to 'default'"""
        if "speaker" in item and "language" in item:
            return item
        speaker = "default" if "speaker" not in item else item["speaker"]
        language = "default" if "language" not in item else item["language"]
        return {**item, **{"speaker": speaker, "language": language}}

    def compute_stats(
        self, energy=True, pitch=True
    ) -> tuple[Optional[Scaler], Optional[Scaler]]:
        if self.cpus > 1:
            parallel = Parallel(n_jobs=self.cpus, backend="loky", batch_size=500)
        if energy:
            energy_scaler = Scaler()
            # Until Python 3.13, pathlib.Path.glob() doesn't work with symlinks: https://github.com/python/cpython/issues/77609
            paths = glob(
                str(self.config.preprocessing.save_dir / "energy/**/*energy*"),
                recursive=True,
            )
            if self.cpus > 1:
                logger.info("Gathering energy values")
                with tqdm_joblib_context(tqdm(desc="Gathering energy values")):
                    for energy_data in parallel(
                        delayed(torch.load)(path) for path in paths
                    ):
                        energy_scaler.data.append(energy_data)
            else:
                for path in tqdm(paths, desc="Gathering energy values"):
                    energy_data = torch.load(path)
                    energy_scaler.data.append(energy_data)
        if pitch:
            pitch_scaler = Scaler()
            # Until Python 3.13, pathlib.Path.glob() doesn't work with symlinks: https://github.com/python/cpython/issues/77609
            paths = glob(
                str(self.config.preprocessing.save_dir / "pitch/**/*pitch*"),
                recursive=True,
            )
            if self.cpus > 1:
                logger.info("Gathering pitch values")
                with tqdm_joblib_context(tqdm(desc="Gathering pitch values")):
                    for pitch_data in parallel(
                        delayed(torch.load)(path) for path in paths
                    ):
                        pitch_scaler.data.append(pitch_data)
            else:
                for path in tqdm(paths, desc="Gathering pitch values"):
                    pitch_data = torch.load(path)
                    pitch_scaler.data.append(pitch_data)
        return energy_scaler if energy else energy, pitch_scaler if pitch else pitch

    def normalize_stats(self, energy_scaler: Scaler, pitch_scaler: Scaler):
        """Normalize pitch and energy to unit variance"""
        # Note: this function is IO bound, because it is a tight loop writing small files.
        # Attempts to parallelize it make it much slower, even with only 2 threads.
        logger.info("Scaling dataset statistics...")
        stats = {}
        if energy_scaler:
            energy_stats = energy_scaler.calculate_stats()
            for path in tqdm(
                # Until Python 3.13, pathlib.Path.glob() doesn't work with symlinks: https://github.com/python/cpython/issues/77609
                glob(
                    str(self.config.preprocessing.save_dir / "energy/**/*energy*"),
                    recursive=True,
                ),
                desc="Normalizing energy values",
            ):
                energy = torch.load(path)
                energy = energy_scaler.normalize(energy)
                save_tensor(energy, path)
            stats["energy"] = energy_stats
        if pitch_scaler:
            pitch_stats = pitch_scaler.calculate_stats()
            for path in tqdm(
                # Until Python 3.13, pathlib.Path.glob() doesn't work with symlinks: https://github.com/python/cpython/issues/77609
                glob(
                    str(self.config.preprocessing.save_dir / "pitch/**/*pitch*"),
                    recursive=True,
                ),
                desc="Normalizing pitch values",
            ):
                pitch = torch.load(path)
                pitch = pitch_scaler.normalize(pitch)
                save_tensor(pitch, path)
            stats["pitch"] = pitch_stats

        return stats

    def dataset_sanity_checks(self):
        """Before processing datasets, we should do some sanity checks."""
        for dataset in self.datasets:
            data_dir = Path(dataset.data_dir)
            if not data_dir.exists():
                logger.error(
                    f"Data directory '{data_dir}' does not exist. Please check your config file."
                )
                sys.exit(1)

    def create_path(self, item: dict, folder: str, fn: str) -> Path:
        path = (
            self.save_dir
            / folder
            / self.sep.join([item["basename"], item["speaker"], item["language"], fn])
        )
        return path

    def process_one_audio(
        self, item: dict, data_dir, sox_effects: list[list]
    ) -> Optional[dict]:
        """Process one audio item

        Return:
           - item if it is found and processed successfully
           - None otherwise, indicating it should be skipped from further processing
        """
        extension = "" if item["basename"].endswith(".wav") else ".wav"
        audio_path = data_dir / (item["basename"] + extension)
        if not audio_path.exists():
            logger.warning(f"File '{item}' is missing and will not be processed.")
            self.counters.increment("missing_files")
            return None

        item = self.get_speaker_and_language(item)
        input_audio_save_path = self.create_path(
            item, "audio", f"audio-{self.input_sampling_rate}.wav"
        )
        output_audio_save_path = self.create_path(
            item, "audio", f"audio-{self.output_sampling_rate}.wav"
        )
        if (
            input_audio_save_path.exists()
            and output_audio_save_path.exists()
            and not self.overwrite
        ):
            audio, sr, seconds = self.load_audio(audio_path)
            self.counters.increment("previously_processed_files")
            self.counters.increment("duration", seconds)
            return item
        if not input_audio_save_path.exists() or self.overwrite:
            input_audio, save_sr = self.process_audio(
                audio_path,
                resample_rate=self.input_sampling_rate,
                sox_effects=sox_effects,
                hop_size=self.audio_config.fft_hop_size,
            )
            if input_audio is None:
                return None
            else:
                input_audio = input_audio.unsqueeze(0)
                save_wav(
                    input_audio,
                    input_audio_save_path,
                    save_sr,
                    self.audio_config.target_bit_depth,
                )

        if (
            self.input_sampling_rate != self.output_sampling_rate
            and not output_audio_save_path.exists()
            or self.overwrite
        ):
            output_audio, save_sr = self.process_audio(
                audio_path,
                resample_rate=self.output_sampling_rate,
                sox_effects=sox_effects,
                update_counters=False,
                hop_size=self.output_hop_size,
            )
            if output_audio is not None:
                output_audio = output_audio.unsqueeze(0)
                save_wav(
                    output_audio,
                    input_audio_save_path,
                    save_sr,
                    self.audio_config.target_bit_depth,
                )
        return item

    def process_all_audio(self) -> list[dict]:
        """Process all audio across datasets, create a combined, filtered filelist and return it"""
        self.dataset_sanity_checks()
        filtered_filelist: list[dict] = []
        for dataset in tqdm(self.datasets, total=len(self.datasets), desc="Dataset"):
            data_dir = Path(dataset.data_dir)
            filelist = dataset.filelist_loader(dataset.filelist)
            sox_effects = dataset.sox_effects
            if self.debug:
                filelist = filelist[:10]
                logger.info(
                    "Debug flag was set to true, only processing first 10 files"
                )
            logger.info(f"Collecting files for {dataset.label} and processing audio")
            if self.cpus * 3 > len(filelist):
                self.cpus = len(filelist) // 3
            if self.cpus > 1:
                logger.info("Launching parallel processes may take a moment...")
                batch_size = min(100, 1 + len(filelist) // (self.cpus * 2))
                with tqdm_joblib_context(
                    tqdm(
                        desc=f"Processing audio on {self.cpus} CPUs",
                        total=len(filelist),
                    )
                ):
                    processed_items = Parallel(
                        n_jobs=self.cpus,
                        # verbose=10,
                        backend="loky",
                        batch_size=batch_size,
                    )(
                        delayed(self.process_one_audio)(item, data_dir, sox_effects)
                        for item in filelist
                    )
                filtered_filelist.extend(
                    item for item in processed_items if item is not None
                )
            else:
                for item in tqdm(filelist, desc="Processing Audio on 1 CPU"):
                    processed_item = self.process_one_audio(item, data_dir, sox_effects)
                    if processed_item is not None:
                        filtered_filelist.append(processed_item)
        return filtered_filelist

    def process_energy(self, item):
        energy_path = self.create_path(item, "energy", "energy.pt")
        # Always reprocess energy even without self.overwrite, since its results
        # depend on the stats of the whole fileset.
        spec_path = self.create_path(
            item,
            "spec",
            f"spec-{self.input_sampling_rate}-{self.audio_config.spec_type}.pt",
        )
        spec = torch.load(spec_path)
        energy = self.extract_energy(spec)
        if (
            isinstance(self.config, FeaturePredictionConfig)
            and self.config.model.variance_predictors.energy.level == "phone"
            and not self.config.model.learn_alignment
        ):
            dur_path = self.create_path(item, "duration", "duration.pt")
            durs = torch.load(dur_path)
            energy = self.average_data_by_durations(energy, durs)
        save_tensor(energy, energy_path)

    def process_pitch(self, item):
        pitch_path = self.create_path(item, "pitch", "pitch.pt")
        # Always reprocess pitch even without self.overwrite, since its results
        # depend on the stats of the whole fileset.
        audio_path = self.create_path(
            item, "audio", f"audio-{self.input_sampling_rate}.wav"
        )
        audio, _, _ = self.load_audio(audio_path)
        pitch = self.extract_pitch(audio)
        if (
            isinstance(self.config, FeaturePredictionConfig)
            and self.config.model.variance_predictors.pitch.level == "phone"
            and not self.config.model.learn_alignment
        ):
            dur_path = self.create_path(item, "duration", "duration.pt")
            durs = torch.load(dur_path)
            pitch = self.average_data_by_durations(pitch, durs)
        save_tensor(pitch, pitch_path)

    def process_attn_prior(self, item):
        if self.text_processor is None:
            raise NotImplementedError(
                "You must have a valid TextProcessor in order to calculate the attention prior."
            )
        # Create an attention prior for characters and one for phones if available
        process_characters = False
        process_phones = False
        character_tokens = (
            self.text_processor.encode_string_tokens(
                item["character_tokens"].split("/")
            )
            if "character_tokens" in item
            and item[
                "character_tokens"
            ]  # character_tokens will be None in phone-based datasets. It will be the empty string in multi-source datasets which contain characters for other datasets.
            else None
        )
        phone_tokens = (
            self.text_processor.encode_string_tokens(item["phone_tokens"].split("/"))
            if "phone_tokens" in item and item["phone_tokens"]
            else None
        )
        if character_tokens:
            character_attn_prior_path = self.create_path(
                item,
                "attn",
                f"{DatasetTextRepresentation.characters.value}-attn-prior.pt",
            )
            process_characters = True
        if phone_tokens:
            phone_attn_prior_path = self.create_path(
                item,
                "attn",
                f"{DatasetTextRepresentation.ipa_phones.value}-attn-prior.pt",
            )
            process_phones = True
        if (
            process_characters
            and character_attn_prior_path.exists()
            and not self.overwrite
        ):
            process_characters = False
        if process_phones and phone_attn_prior_path.exists() and not self.overwrite:
            process_phones = False
        if not process_phones and not process_characters:
            return
        binomial_interpolator = BetaBinomialInterpolator()

        input_spec_path = self.create_path(
            item,
            "spec",
            f"spec-{self.input_sampling_rate}-{self.audio_config.spec_type}.pt",
        )
        input_spec = torch.load(input_spec_path)
        if process_phones:
            phone_attn_prior = torch.from_numpy(
                binomial_interpolator(input_spec.size(1), len(phone_tokens))
            )
            assert input_spec.size(1) == phone_attn_prior.size(0)
            assert len(phone_tokens) == phone_attn_prior.size(1)
            save_tensor(phone_attn_prior, phone_attn_prior_path)
        if process_characters:
            character_attn_prior = torch.from_numpy(
                binomial_interpolator(input_spec.size(1), len(character_tokens))
            )
            assert input_spec.size(1) == character_attn_prior.size(0)
            assert len(character_tokens) == character_attn_prior.size(1)
            save_tensor(character_attn_prior, character_attn_prior_path)

    @staticmethod
    def process_text(
        item,
        text_processor: TextProcessor,
        use_pfs=False,
        specific_text_representation: Optional[
            TargetTrainingTextRepresentationLevel
        ] = None,
        encode_as_string=True,
    ) -> (
        tuple[Optional[str], Optional[str], Optional[npt.NDArray[np.float32]]]
        | tuple[
            Optional[list[int]], Optional[list[int]], Optional[npt.NDArray[np.float32]]
        ]
    ):
        """Process text into characters, phones, and/or phonological features.
            If specific_text_representation is supplied, then only that representation is processed.
            Otherwise:
                - if "characters" are defined on the dataset item, they are tokenized and one hot encoded
                    - if a g2p method is available for the defined item's language, then they are also g2p'ed and phones are processed
                - if "phones" are defined on the dataset item, they are tokenized and one-hot encoded
                - if "arpabet" is defined on the dataset item, it is converted to phones and one-hot encoded
                - if "use_pfs" is True and phones are available, a phonological feature representation is also generated

        Args:
            item (_type_): an item from the filelist
            use_pfs (bool, optional): whether to generate phonological features. Defaults to False.
            specific_text_representation (Optional[DatasetTextRepresentation], optional): a specific representation to process if you don't want the method to create all available representations, useful for synthesis. Defaults to None.
            encode_as_string (bool, optional): whether to return characters and phones as /-joined strings (when True) or lists of ints (when False). Defaults to True.

        Returns:
            tuple[Optional[str], Optional[str], Optional[npt.NDArray[np.float32]]]|tuple[Optional[list[int]], Optional[list[int]], Optional[npt.NDArray[np.float32]]]: if encode_as_string is true, returns an optional characters string, an optional phones string, and an optional multi-hot phonological feature vector. if encode_as_string is false, returns a list of ints for characters and phones
        """
        if specific_text_representation is not None:
            raise NotImplementedError(
                "Sorry 'specific_text_representation' isn't implemented yet, please set it to None."
            )  # TODO: refactor so that you don't *need* to generate all possible representations, to make synthesis faster.
        if text_processor is None:
            raise NotImplementedError(
                "You must have a valid TextProcessor in order to calculate text."
            )
        characters: Optional[str] = None
        phones: Optional[str] = None
        phone_tokens: Optional[list[int]] = None
        character_tokens: Optional[list[int]] = None
        pfs: Optional[npt.NDArray[np.float32]] = None
        # if arpabet, turn to phones right away
        if (
            DatasetTextRepresentation.arpabet.value in item
            and DatasetTextRepresentation.ipa_phones.value not in item
        ):
            tokens = text_processor.encode_text(
                text=ARPABET_TO_IPA_TRANSDUCER(
                    item[DatasetTextRepresentation.arpabet.value]
                ).output_string,
                quiet=True,
                apply_g2p=False,
            )
            assert isinstance(tokens, list)
            phone_tokens = tokens
        # if dataset is chars, tokenize them for saving to filelist
        if DatasetTextRepresentation.characters.value in item:
            tokens = text_processor.encode_text(
                text=item[DatasetTextRepresentation.characters.value],
                apply_g2p=False,
                encode_as_phonological_features=False,
                quiet=True,
            )
            assert isinstance(tokens, list)
            character_tokens = tokens

            # if g2p available, and phones don't already exist, process and tokenize them for saving to filelist
            if (
                item["language"] in AVAILABLE_G2P_ENGINES
                and DatasetTextRepresentation.ipa_phones.value not in item
            ):
                tokens = text_processor.encode_text(
                    text=item[DatasetTextRepresentation.characters.value],
                    apply_g2p=True,
                    lang_id=item["language"],
                    encode_as_phonological_features=False,
                    quiet=True,
                )
                assert isinstance(tokens, list)
                phone_tokens = tokens
        # if dataset is phones
        if DatasetTextRepresentation.ipa_phones.value in item:
            tokens = text_processor.encode_text(
                text=item[DatasetTextRepresentation.ipa_phones.value],
                apply_g2p=False,
                encode_as_phonological_features=False,
                quiet=True,
            )
            assert isinstance(tokens, list)
            phone_tokens = tokens
        # calculate pfs
        if phone_tokens and use_pfs:
            pfs = text_processor.calculate_phonological_features(
                text_processor.token_sequence_to_text_sequence(phone_tokens),
                apply_punctuation_rules=True,
            )
        # encode to string
        if encode_as_string:
            if phone_tokens is not None:
                phones = text_processor.decode_tokens(phone_tokens, join_character="/")
            if character_tokens is not None:
                characters = text_processor.decode_tokens(
                    character_tokens, join_character="/"
                )
            return (characters, phones, pfs)
        else:
            return (character_tokens, phone_tokens, pfs)

    def process_spec(
        self, item
    ) -> tuple[Optional[torch.Tensor], Optional[torch.Tensor]]:
        """Processes spectral features based on the defined transform (linear, Mel, complex etc)
        Processes an 'input' spectrogram based on the sampling rate of the audio input to the vocoder.
        Processes an 'output' spectrogram based on the output sampling rate of the vocoder.

        We also limit the length of the audio to a number that is evenly divisible by the hop size
        """
        input_spec = None
        output_spec = None
        input_audio_path = self.create_path(
            item, "audio", f"audio-{self.input_sampling_rate}.wav"
        )
        if not input_audio_path.exists():
            self.counters.increment("skipped_processes")
            logger.info(f"Audio at {input_audio_path} is missing. Skipping...")
            return input_spec, output_spec
        output_audio_path = self.create_path(
            item, "audio", f"audio-{self.output_sampling_rate}.wav"
        )
        input_spec_path = self.create_path(
            item,
            "spec",
            f"spec-{self.input_sampling_rate}-{self.audio_config.spec_type}.pt",
        )

        if input_audio_path != output_audio_path:
            output_spec_path = self.create_path(
                item,
                "spec",
                f"spec-{self.output_sampling_rate}-{self.audio_config.spec_type}.pt",
            )
            if not output_spec_path.exists() or self.overwrite:
                output_audio, _, _ = self.load_audio(output_audio_path)
                output_audio = output_audio.squeeze()
                # limit the number of frames to be equal to the number of
                # available full frames from the audio (equal to fft_hop_size * sampling rate change, if we are upsampling)
                # i.e. we round down
                max_output_frames = output_audio.size(0) // self.output_hop_size
                output_spec = self.extract_spectral_features(
                    output_audio, self.output_spectral_transform
                )[:, :max_output_frames]
                assert max_output_frames == output_spec.size(1)
                save_tensor(output_spec, output_spec_path)

        if not input_spec_path.exists() or self.overwrite:
            input_audio, _, _ = self.load_audio(input_audio_path)
            input_audio = input_audio.squeeze()
            # limit the number of frames to be equal to the number of
            # available full frames from the audio
            # i.e. we round down
            max_input_frames = input_audio.size(0) // self.audio_config.fft_hop_size
            input_spec = self.extract_spectral_features(
                input_audio, self.input_spectral_transform
            )[:, :max_input_frames]
            assert max_input_frames == input_spec.size(1)
            save_tensor(input_spec, input_spec_path)
        return input_spec, output_spec

    def get_process_fn(self, process):
        if process == "text":
            return functools.partial(
                self.process_text, text_processor=self.text_processor, use_pfs=False
            )
        if process == "pfs":
            return functools.partial(
                self.process_text, text_processor=self.text_processor, use_pfs=True
            )
        if process == "energy":
            return self.process_energy
        if process == "pitch":
            return self.process_pitch
        if process == "spec":
            return functools.partial(self.process_spec)
        if process == "attn":
            return self.process_attn_prior

    def check_data(self, filelist, word_seg_token=" ", heavy_clip_detction=False):
        data = []
        # speaking rate (words/second, float, scatterplot or bar chart)
        # speaking rate (characters/second, float, scatterplot or bar chart)
        # articulation level (mean energy/speaking rate)
        # unrecognized symbols (bool, list)
        # duration (float, box plot)
        # clipping (float, box plot)
        # silence % (float, box plot)
        for item in tqdm(filelist, desc="Checking Data"):
            data_point = {k: v for k, v in item.items()}
            raw_text = item["text"]
            n_words = len(raw_text.split(word_seg_token))
            n_chars = len(self.text_processor.encode_text(raw_text))
            audio, _, _ = torchaudio.load(
                self.create_path(item, "audio", f"audio-{self.input_sampling_rate}.wav")
            )
            if heavy_clip_detction:
                _, total_clipping = detect_clipping(audio)
            else:
                # this isn't a great way of detecting clipping,
                # but it's a lot faster than clipdetect
                audio_max = audio.max()
                audio_min = audio.min()
                total_clipping = (
                    audio[audio >= audio_max].size(0)
                    + audio[audio <= audio_min].size(0)
                    - 2
                )
            pitch = torch.load(self.create_path(item, "pitch", "pitch.pt"))
            energy = torch.load(self.create_path(item, "energy", "energy.pt"))
            audio_length_s = len(audio) / self.input_sampling_rate
            data_point["total_clipped_samples"] = total_clipping
            data_point["pitch_min"] = float(pitch.min())
            data_point["pitch_max"] = float(pitch.max())
            data_point["pitch_mean"] = float(pitch.mean())
            data_point["pitch_std"] = float(pitch.std())
            data_point["energy_min"] = float(energy.min())
            data_point["energy_max"] = float(energy.max())
            data_point["energy_mean"] = float(energy.mean())
            data_point["energy_std"] = float(energy.std())
            data_point["duration"] = audio_length_s
            data_point["speaking_rate_word"] = n_words / audio_length_s
            data_point["speaking_rate_char"] = n_chars / audio_length_s
            data_point["articulation_rate_word"] = (
                data_point["energy_mean"] / data_point["speaking_rate_word"]
            )
            data_point["articulation_rate_char"] = (
                data_point["energy_mean"] / data_point["speaking_rate_char"]
            )
            data_point["n_missing_symbols"] = len(
                self.text_processor.get_missing_symbols(raw_text)
            )
            data_point["n_words"] = n_words
            data_point["n_chars"] = n_chars
            data.append(data_point)
        return data

    def load_filelist(self, path: Path):
        try:
            filelist = generic_psv_filelist_reader(path)
            if self.debug:
                logger.info(
                    "Debug flag was set to true, only processing first 10 files"
                )
                filelist = filelist[:10]
        except FileNotFoundError:
            logger.error(
                f"A filelist was not found at {path}. "
                "Please try processing your audio again."
            )
            sys.exit(1)
        return filelist

    def get_config_lock(self, in_progress: bool = True) -> dict[str, dict | str]:
        """Get the part of the configuration in use that affects preprocessing output"""

        # Every config class has .preprocessing.audio as a dict
        # Must be identical between processing runs
        config_audio = self.config.preprocessing.audio.model_dump(mode="json")

        # Every config class has .preprocessing.source_data as a list of dicts
        # We use each source's label as key
        # For each entry in common between the new and old config, the entry
        # details must be the same
        source_data_dict = {
            entry.label: entry.model_dump(mode="json", exclude={"data_dir", "filelist"})
            for entry in self.config.preprocessing.source_data
        }

        # Not every config class has .text (e.g., VocoderConfig doesn't)
        # When found, this must be identical between preprocessing runs
        if hasattr(self.config, "text"):
            config_text = self.config.text.model_dump(mode="json")
        else:
            config_text = {}

        return {
            "info": "This file has the configuration that was used to preprocess files. Do not edit.",
            "status": "in progress" if in_progress else "completed",
            "preprocessing.audio": config_audio,
            "preprocessing.source_data": source_data_dict,
            "text": config_text,
        }

    def save_config_lock(self, in_progress: bool):
        """Write the part of the configuration in use that affects preprocessing output

        Args:
            in_progress: indicates if preprocessing is in progress or has completed
        """
        self.save_dir.mkdir(parents=True, exist_ok=True)
        lock_file = self.save_dir / ".config-lock"
        if lock_file.exists():
            lock_file.chmod(0o666)  # make it writable to overwrite it
        with open(lock_file, "w", encoding="utf8") as f:
            json.dump(
                self.get_config_lock(in_progress), f, indent=2, ensure_ascii=False
            )
            f.write("\n")
        lock_file.chmod(0o444)  # leave it read-only so users to discourage change

    def load_config_lock(self):
        """Load the existing config lock, returning it parsed, or None if not found"""
        lock_file = self.save_dir / ".config-lock"
        try:
            with open(lock_file, "r", encoding="utf8") as f:
                return json.load(f)
        except FileNotFoundError:
            return None
        except json.JSONDecodeError as e:
            logger.warning(
                f"Error loading existing config lock {lock_file}: {e}. Assuming a configuration mismatch."
            )
            return {
                "preprocessing.audio": "ERROR: Unable to load",
                "text": "ERROR: Unable to load",
            }

    def config_lock_has_conflicts(self) -> bool:
        current_config_lock = self.get_config_lock()
        saved_config_lock = self.load_config_lock()
        if saved_config_lock is None:
            return False  # No existing config lock found means no conflicts

        found_conflict = False

        # preprocessing that was interrupted in progress cannot be trusted
        if saved_config_lock.get("status") != "completed":
            logger.warning(
                "The Config lock indicates that the previous preprocessing run was interrupted in progress. We won't trust these partial results."
            )
            found_conflict = True

        # preprocessing.audio config has to be identical
        cur_audio = current_config_lock["preprocessing.audio"]
        locked_audio = saved_config_lock.get("preprocessing.audio")
        if cur_audio != locked_audio:
            logger.warning(
                f"Config lock mismatch: current preprocessing.audio config:\n{cur_audio}\ndiffers from locked preprocessing.audio config:\n{locked_audio}"
            )
            found_conflict = True

        # text config has to be identical
        cur_text = current_config_lock["text"]
        locked_text = saved_config_lock.get("text")
        if cur_text != locked_text:
            logger.warning(
                f"Config lock mismatch: current text config:\n{cur_text}\ndiffers from locked text config:\n{locked_text}"
            )
            found_conflict = True

        # preprocessing.source_data entries in common before and after must be equal
        cur_source_data = current_config_lock["preprocessing.source_data"]
        locked_source_data = saved_config_lock.get("preprocessing.source_data", {})
        for label in cur_source_data.keys() & locked_source_data.keys():  # type: ignore[union-attr]
            if cur_source_data[label] != locked_source_data[label]:
                logger.warning(
                    f"Config lock mismatch: current preprocessing.source_data[{label}] config:\n{cur_source_data[label]}\ndiffers from locked preprocessing.source_data[{label}] config:\n{locked_source_data[label]}"
                )
                found_conflict = True

        return found_conflict

    def preprocess(  # noqa: C901
        self,
        output_path="filelist.psv",
        cpus=min(5, mp.cpu_count()),
        to_process=list[str],
        overwrite=False,
        debug=False,
    ):
        self.overwrite = overwrite

        if not overwrite:
            if self.config_lock_has_conflicts():
                logger.error(
                    "Config lock mismatch: you have previously preprocessed some of the files here, but with a different and potentially incompatible configuration. Please use --overwrite to reprocess all files."
                )
                sys.exit(1)
        self.save_config_lock(in_progress=True)

        self.cpus = cpus
        self.debug = debug
        if not isinstance(output_path, Path):
            output_path = Path(output_path)
        processing_order = ("audio", "text", "pfs", "spec", "attn", "energy", "pitch")
        random.seed(self.config.preprocessing.dataset_split_seed)
        processed_filelist = self.save_dir / output_path.name
        for process in processing_order:
            if process not in to_process:
                continue
            if (
                process != "text"
            ):  # text gets stored in the filelist directly and not to disk
                (self.save_dir / process).mkdir(parents=True, exist_ok=True)
            if process == "audio":
                if filelist := self.process_all_audio():
                    write_filelist(filelist, processed_filelist)
                    report = self.report()
                    with open(self.save_dir / "summary.txt", "w", encoding="utf8") as f:
                        f.write(report)
                    rich_print(report)
                else:
                    logger.error(
                        "Your filtered audio filelist is empty. Nothing to process."
                    )
                    sys.exit(1)
                # logger.info(f"Audio Filelist len={len(filelist or [])}")
            elif process in ["text", "pfs"]:
                # We split out the "text" step to issue the missing symbol warnings
                filelist = self.load_filelist(processed_filelist)
                process_fn = self.get_process_fn(process)
                missing_symbols_before = Counter(self.text_processor.missing_symbols)
                for i, f in tqdm(
                    enumerate(filelist), desc=f"Processing {process} on 1 CPU"
                ):
                    characters, phones, pfs = process_fn(f)
                    if characters is not None:
                        filelist[i]["character_tokens"] = characters
                    if phones is not None:
                        filelist[i]["phone_tokens"] = phones
                    if pfs is not None:
                        pfs_path = self.create_path(f, "pfs", "pfs.pt")
                        pfs_tensor = torch.from_numpy(pfs)
                        torch.save(pfs_tensor, pfs_path)
                # Write filelist after with character and phone tokens
                write_filelist(filelist, processed_filelist)
                # if only one of "pfs" or "text" is specified, missing_symbols_before
                # will always be empty, but if both are specified this makes sure
                # each process gets only its own missing symbols logged.
                new_missing_symbols = (
                    self.text_processor.missing_symbols - missing_symbols_before
                )
                for symbol, count in new_missing_symbols.items():
                    logger.warning(
                        f"Symbol '{symbol}' occurs {n_times(count)} but was not declared in your configuration so it is being ignored."
                    )
            else:
                # If audio has already been processed, then just read the processed_filelist
                filelist = self.load_filelist(processed_filelist)
                process_fn = self.get_process_fn(process)
                logger.info(f"Processing {process} on {self.cpus} CPUs...")
                # logger.info(f"Filelist len={len(filelist or [])}")
                if self.cpus > 1:
                    batch_size = min(100, 1 + len(filelist) // (self.cpus * 2))
                    with tqdm_joblib_context(
                        tqdm(
                            desc=f"Processing {process} on {self.cpus} CPUs",
                            total=len(filelist),
                        )
                    ):
                        Parallel(
                            n_jobs=self.cpus,
                            backend="loky",
                            batch_size=batch_size,
                        )(delayed(process_fn)(file) for file in filelist)
                else:
                    for f in tqdm(filelist, desc=f"Processing {process} on 1 CPU"):
                        process_fn(f)
        # re-load the filelist, sample the validation set and subtract it from the whole dataset to determine the training set
        filelist = self.load_filelist(processed_filelist)
        random.shuffle(filelist)
        train_split = int(len(filelist) * self.config.preprocessing.train_split)
        write_filelist(
            filelist[:train_split],
            self.save_dir / f"training_{output_path.name}",
        )
        write_filelist(
            filelist[train_split:],
            self.save_dir / f"validation_{output_path.name}",
        )
        if "audio" in to_process:
            report = f"Here is a report:\n {self.report()}"
            if not self.counters.value("duration"):
                report += "\n\nWARNING: No audio files were processed."
        else:
            report = ""
        rich_print(
            Panel(
                f"You've finished preprocessing: {', '.join(to_process)}. Your files are located at {self.save_dir.absolute()}. {report}",
                title="Congratulations 🎉",
                subtitle="Next Steps Documentation: https://roedoejet.github.io/EveryVoice/guides/custom.html",
                border_style=Style(color="#0B4F19"),
            )
        )

        self.save_config_lock(in_progress=False)



# flake8: noqa
"""These are function stubs whose sole purpose is to merge with the function signatures of model-specific
    cli command functions. Each of these should have a typer default (which can be overidden by the model-specific command)
    there should be no body as the union of these signatures and the model-specific signatures is what the helper function
    will be called with.
"""
import multiprocessing as mp
from pathlib import Path
from typing import List, Optional

import typer


def complete_path() -> list[str]:
    # https://github.com/tiangolo/typer/discussions/625
    # Work-around for path completion bug in CLI autocompletion
    return []


def load_config_base_command_interface(
    config_file: Path = typer.Argument(
        ...,
        exists=True,
        dir_okay=False,
        file_okay=True,
        help="The path to your model configuration file.",
        autocompletion=complete_path,
    ),
    config_args: List[str] = typer.Option(None, "--config", "-c"),
):
    pass


def preprocess_base_command_interface(
    config_file: Path = typer.Argument(
        ...,
        exists=True,
        dir_okay=False,
        file_okay=True,
        help="The path to your model configuration file.",
        autocompletion=complete_path,
    ),
    config_args: List[str] = typer.Option(
        None, "-c", "--config-args", help="Override the configuration."
    ),
    cpus: Optional[int] = typer.Option(
        min(4, mp.cpu_count()),
        "-C",
        "--cpus",
        help="How many CPUs to use when preprocessing",
    ),
    overwrite: bool = typer.Option(
        False,
        "-O",
        "--overwrite",
        help="Redo all preprocessing, even if files already exist and aren't expected to change.",
    ),
    debug: bool = typer.Option(False, "-D", "--debug", help="Enable debugging."),
):
    pass


def train_base_command_interface(
    config_file: Path = typer.Argument(
        ...,
        exists=True,
        dir_okay=False,
        file_okay=True,
        help="The path to your model configuration file.",
        autocompletion=complete_path,
    ),
    config_args: List[str] = typer.Option(
        None, "-c", "--config-args", help="Overwrite the configuration"
    ),
    accelerator: str = typer.Option(
        "auto",
        "--accelerator",
        "-a",
        help="Uses PyTorch Lightning Accelerators: https://pytorch-lightning.readthedocs.io/en/stable/extensions/accelerator.html",
    ),
    devices: str = typer.Option(
        "auto", "--devices", "-d", help="The number of GPUs on each node"
    ),
    nodes: int = typer.Option(
        1, "--nodes", "-n", help="The number of nodes on your machine"
    ),
    strategy: str = typer.Option(
        "ddp",
        "--strategy",
        "-s",
        help="The strategy for data parallelization: https://pytorch-lightning.readthedocs.io/en/stable/accelerators/gpu_intermediate.html",
    ),
):
    pass


def inference_base_command_interface(
    config_file: Path = typer.Argument(
        ...,
        exists=True,
        dir_okay=False,
        file_okay=True,
        help="The path to your model configuration file.",
        autocompletion=complete_path,
    ),
    config_args: List[str] = typer.Option(None, "--config", "-c"),
    accelerator: str = typer.Option(
        "auto",
        "--accelerator",
        "-a",
        help="Uses PyTorch Lightning Accelerators: https://pytorch-lightning.readthedocs.io/en/stable/extensions/accelerator.html",
    ),
    devices: str = typer.Option(
        "auto", "--devices", "-d", help="The number of GPUs on each node"
    ),
    nodes: int = typer.Option(
        1, "--nodes", "-n", help="The number of nodes on your machine"
    ),
    strategy: str = typer.Option(
        "ddp",
        "--strategy",
        "-s",
        help="The strategy for data parallelization: https://pytorch-lightning.readthedocs.io/en/stable/accelerators/gpu_intermediate.html",
    ),
):
    pass
